# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
import pandas as pd
import scipy.sparse as sp
import GridCal.Engine.Core.topology as tp


def get_bus_indices(C_branch_bus):
    F = np.zeros(C_branch_bus.shape[0], dtype=int)

    for j in range(C_branch_bus.shape[1]):
        for l in range(C_branch_bus.indptr[j], C_branch_bus.indptr[j + 1]):
            i = C_branch_bus.indices[l]  # row index
            F[i] = j

    return F


class BranchData:

    def __init__(self, nbr, nbus, ntime=1):
        """
        Branch data arrays
        :param nbr: number of branches
        :param nbus: number of buses
        """
        self.nbr = nbr
        self.ntime = ntime

        self.branch_names = np.empty(self.nbr, dtype=object)

        self.branch_dc = np.zeros(self.nbr, dtype=int)

        self.branch_active = np.zeros((nbr, ntime), dtype=int)
        self.branch_rates = np.zeros((nbr, ntime), dtype=float)
        self.branch_contingency_rates = np.zeros((nbr, ntime), dtype=float)

        self.F = np.zeros(self.nbr, dtype=int)  # indices of the "from" buses
        self.T = np.zeros(self.nbr, dtype=int)  # indices of the "to" buses

        # composite losses curve (a * x^2 + b * x + c)
        self.a = np.zeros(self.nbr, dtype=float)
        self.b = np.zeros(self.nbr, dtype=float)
        self.c = np.zeros(self.nbr, dtype=float)

        self.R = np.zeros(self.nbr, dtype=float)
        self.X = np.zeros(self.nbr, dtype=float)
        self.G = np.zeros(self.nbr, dtype=float)
        self.B = np.zeros(self.nbr, dtype=float)
        self.k = np.ones(nbr, dtype=float)

        self.m = np.ones((nbr, ntime), dtype=float)
        self.m_min = 0.1 * np.ones(nbr, dtype=float)
        self.m_max = 1.5 * np.ones(nbr, dtype=float)
        self.theta = np.zeros((nbr, ntime), dtype=float)
        self.theta_min = - 6.28 * np.ones(nbr, dtype=float)
        self.theta_max = 6.28 * np.ones(nbr, dtype=float)
        self.Beq = np.zeros((nbr, ntime), dtype=float)
        self.G0 = np.zeros((nbr, ntime), dtype=float)

        self.tap_t = np.ones(self.nbr, dtype=float)
        self.tap_f = np.ones(self.nbr, dtype=float)

        self.Pfset = np.zeros((nbr, ntime))
        self.Qfset = np.zeros((nbr, ntime))
        self.Qtset = np.zeros((nbr, ntime))
        self.vf_set = np.ones((nbr, ntime))
        self.vt_set = np.ones((nbr, ntime))

        self.Kdp = np.ones(self.nbr)
        self.Kdp_va = np.ones(self.nbr)
        self.alpha1 = np.zeros(self.nbr)  # converter losses parameter (alpha1)
        self.alpha2 = np.zeros(self.nbr)  # converter losses parameter (alpha2)
        self.alpha3 = np.zeros(self.nbr)  # converter losses parameter (alpha3)
        self.control_mode = np.zeros(self.nbr, dtype=object)

        self.contingency_enabled = np.ones(self.nbr, dtype=int)
        self.monitor_loading = np.ones(self.nbr, dtype=int)

        self.C_branch_bus_f = sp.lil_matrix((self.nbr, nbus), dtype=int)  # connectivity branch with their "from" bus
        self.C_branch_bus_t = sp.lil_matrix((self.nbr, nbus), dtype=int)  # connectivity branch with their "to" bus

    def slice(self, elm_idx, bus_idx, time_idx=None):
        """
        Slice this class
        :param elm_idx: branch indices
        :param bus_idx: bus indices
        :param time_idx: array of time indices
        :return: new BranchData instance
        """

        if time_idx is None:
            tidx = elm_idx
        else:
            tidx = np.ix_(elm_idx, time_idx)

        data = BranchData(nbr=len(elm_idx), nbus=len(bus_idx))

        data.branch_names = self.branch_names[elm_idx]
        # data.F = self.F[elm_idx]
        # data.T = self.T[elm_idx]
        data.R = self.R[elm_idx]
        data.X = self.X[elm_idx]
        data.G = self.G[elm_idx]
        data.B = self.B[elm_idx]
        data.k = self.k[elm_idx]
        data.tap_t = self.tap_f[elm_idx]
        data.tap_f = self.tap_t[elm_idx]
        data.Kdp = self.Kdp[elm_idx]
        data.Kdp_va = self.Kdp_va[elm_idx]
        data.branch_dc = self.branch_dc[elm_idx]
        data.alpha1 = self.alpha1[elm_idx]
        data.alpha2 = self.alpha2[elm_idx]
        data.alpha3 = self.alpha3[elm_idx]

        data.control_mode = self.control_mode[elm_idx]
        data.contingency_enabled = self.contingency_enabled[elm_idx]
        data.monitor_loading = self.monitor_loading[elm_idx]

        data.branch_active = self.branch_active[tidx]
        data.branch_rates = self.branch_rates[tidx]
        data.branch_contingency_rates = self.branch_contingency_rates[tidx]
        data.m = self.m[tidx]

        data.m_min = self.m_min[elm_idx]
        data.m_max = self.m_max[elm_idx]
        data.theta = self.theta[tidx]
        data.theta_min = self.theta_min[elm_idx]
        data.theta_max = self.theta_max[elm_idx]
        data.Beq = self.Beq[tidx]
        data.G0 = self.G0[tidx]
        data.Pfset = self.Pfset[tidx]
        data.Qfset = self.Qfset[tidx]
        data.Qtset = self.Qtset[tidx]
        data.vf_set = self.vf_set[tidx]
        data.vt_set = self.vt_set[tidx]

        data.C_branch_bus_f = self.C_branch_bus_f[np.ix_(elm_idx, bus_idx)]
        data.C_branch_bus_t = self.C_branch_bus_t[np.ix_(elm_idx, bus_idx)]

        data.F = get_bus_indices(data.C_branch_bus_f)
        data.T = get_bus_indices(data.C_branch_bus_t)

        return data

    def get_island(self, bus_idx, t_idx=0):
        """
        get the array of branch indices that belong to the islands given by the bus indices
        :param bus_idx: array of bus indices
        :return: array of island branch indices
        """
        if self.nbr:
            return tp.get_elements_of_the_island(self.C_branch_bus_f + self.C_branch_bus_t,
                                                 island=bus_idx,
                                                 active=self.branch_active[:, t_idx])
        else:
            return np.zeros(0, dtype=int)

    def get_ac_indices(self):
        """

        :return:
        """
        return np.where(self.branch_dc == 0)[0]

    def get_dc_indices(self):
        """

        :return:
        """
        return np.where(self.branch_dc != 0)[0]

    def get_linear_series_admittance(self, t=0):
        """
        Get the linear version of the series admittance for ACDC systems
        :param t: time step index
        :return: Array of the length of the number of branches with 1/X or 1/R depending whether if it is AC or DC
        """
        dc = self.get_dc_indices()
        ac = self.get_ac_indices()
        m_abs = np.abs(self.m[:, t])
        if len(dc):
            # compose the vector for AC-DC grids where the R is needed for this matrix
            # even if conceptually we only want the susceptance
            b = np.zeros(self.nbr)
            active = self.branch_active[:, t]
            b[ac] = 1.0 / (m_abs[ac] * self.X[ac] * active[ac] + 1e-20)  # for ac branches
            b[dc] = 1.0 / (m_abs[dc] * self.R[dc] * active[dc] + 1e-20)  # for dc branches
        else:
            b = 1.0 / (m_abs * self.X * self.branch_active[:, t] + 1e-20)  # for ac branches

        return b

    def get_monitor_enabled_indices(self):
        """

        :return:
        """
        return np.where(self.monitor_loading == 1)[0]

    def get_contingency_enabled_indices(self):
        """

        :return:
        """
        return np.where(self.contingency_enabled == 1)[0]

    def to_df(self, t=0):
        """
        Create DataFrame with the compiled branches information
        :param t: time index, relevant for those magnitudes that change with time
        :return: Pandas DataFrame
        """
        data = {'names': self.branch_names,
                'active': self.branch_active[:, t],
                'F': self.F,
                'T': self.T,
                'Rates': self.branch_rates[:, t],
                'Contingency rates': self.branch_contingency_rates[:, t],
                'R': self.R,
                'X': self.X,
                'G': self.G,
                'B': self.B,
                'Vtap F': self.tap_f,
                'Vtap T': self.tap_t,
                'Tap module': self.m[:, t],
                'Tap angle': self.theta[:, t]}
        return pd.DataFrame(data=data)

    def __len__(self):
        return self.nbr


class BranchOpfData(BranchData):

    def __init__(self, nbr, nbus, ntime=1):
        """

        :param nbr:
        :param nbus:
        :param ntime:
        """
        BranchData.__init__(self, nbr, nbus, ntime)

        self.branch_cost = np.zeros((nbr, ntime), dtype=float)

    def slice(self, elm_idx, bus_idx, time_idx=None):
        """
        Slice this class
        :param elm_idx: branch indices
        :param bus_idx: bus indices
        :param time_idx: array of time indices
        :return: new BranchData instance
        """

        if time_idx is None:
            tidx = elm_idx
        else:
            tidx = np.ix_(elm_idx, time_idx)

        data = super().slice(elm_idx, bus_idx, time_idx)

        data.branch_cost = self.branch_cost[tidx]

        return data




from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager


class ConsoleWidget(RichJupyterWidget):
    """
    Convenience class for a live IPython console widget.
    We can replace the standard banner using the customBanner argument
    """

    def __init__(self, customBanner=None, *args, **kwargs):
        RichJupyterWidget.__init__(self, *args, **kwargs)

        if customBanner is not None:
            self.banner = customBanner

        self.font_size = 6
        self.gui_completion = 'droplist'
        self.kernel_manager = QtInProcessKernelManager()
        self.kernel_manager.start_kernel(show_banner=False)
        self.kernel_manager.kernel.gui = 'qt'
        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()

        def stop():
            kernel_client.stop_channels()
            self.kernel_manager.shutdown_kernel()
            # guisupport.get_app_qt().exit()

        self.exit_requested.connect(stop)

    def push_vars(self, variableDict):
        """
        Given a dictionary containing name / value pairs, push those variables
        to the IPython console widget
        """
        self.kernel_manager.kernel.shell.push(variableDict)

    def clear(self):
        """
        Clears the terminal
        """
        self._control.clear()

        # self.kernel_manager

    def print_text(self, text):
        """
        Prints some plain text to the console
        """
        self._append_plain_text(text)

    def execute_command(self, command):
        """
        Execute a command in the frame of the console widget
        """
        self._execute(command, False)


if __name__ == '__main__':
    # try:
    #     from qtconsole.qt import QtWidgets
    # except:
    from qtconsole.qtconsoleapp import QtWidgets
    app = QtWidgets.QApplication([])
    widget = ConsoleWidget()
    widget.show()
    app.exec_()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from uuid import uuid4
from PySide2.QtCore import QThread, Signal
from typing import List, Dict
import numpy as np

# Module imports
from GridCal.Engine.Simulations.driver_types import SimulationTypes
from GridCal.Engine.Simulations.ATC.available_transfer_capacity_driver import AvailableTransferCapacityResults
from GridCal.Engine.Simulations.ATC.available_transfer_capacity_ts_driver import AvailableTransferCapacityTimeSeriesResults
from GridCal.Engine.Simulations.ContingencyAnalysis.contingency_analysis_results import ContingencyAnalysisResults
from GridCal.Engine.Simulations.ContingencyAnalysis.contingency_analysis_ts_results import ContingencyAnalysisTimeSeriesResults
from GridCal.Engine.Simulations.ContinuationPowerFlow.continuation_power_flow_driver import ContinuationPowerFlowResults
from GridCal.Engine.Simulations.LinearFactors.linear_analysis_driver import LinearAnalysisResults
from GridCal.Engine.Simulations.LinearFactors.linear_analysis_ts_driver import LinearAnalysisTimeSeriesResults
from GridCal.Engine.Simulations.OPF.opf_results import OptimalPowerFlowResults
from GridCal.Engine.Simulations.OPF.opf_ts_results import OptimalPowerFlowTimeSeriesResults
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import PowerFlowResults
from GridCal.Engine.Simulations.PowerFlow.time_series_driver import TimeSeriesResults
from GridCal.Engine.Simulations.ShortCircuitStudies.short_circuit_driver import ShortCircuitResults
from GridCal.Engine.Simulations.Stochastic.stochastic_power_flow_results import StochasticPowerFlowResults
from GridCal.Engine.Simulations.driver_template import DriverTemplate
from GridCal.Engine.Simulations.driver_types import SimulationTypes
from GridCal.Engine.basic_structures import Logger
from GridCal.Gui.Session.results_model import ResultsModel


def get_results_object_dictionary():
    """
    Get dictionary of recognizable result types in order to be able to load a driver from disk
    :return: dictionary[driver name: empty results object]
    """
    lst = [(AvailableTransferCapacityResults([], [], [], []), SimulationTypes.NetTransferCapacity_run),
           (AvailableTransferCapacityTimeSeriesResults([], [], [], [], []), SimulationTypes.NetTransferCapacityTS_run),
           (ContingencyAnalysisResults(0, 0, [], [], []), SimulationTypes.ContingencyAnalysis_run),
           (ContingencyAnalysisTimeSeriesResults(0, 0, 0, [], [], [], []), SimulationTypes.ContingencyAnalysisTS_run),
           (ContinuationPowerFlowResults(0, 0, 0, [], [], []), SimulationTypes.ContinuationPowerFlow_run),
           (LinearAnalysisResults(0, 0, (), (), ()), SimulationTypes.LinearAnalysis_run),
           (LinearAnalysisTimeSeriesResults(0, 0, (), (), (), ()), SimulationTypes.LinearAnalysis_TS_run),
           (OptimalPowerFlowResults((), (), (), (), ()), SimulationTypes.OPF_run),
           (OptimalPowerFlowTimeSeriesResults((), (), (), (), (), (), 0, 0, 0), SimulationTypes.OPFTimeSeries_run),
           (PowerFlowResults(0, 0, 0, 0, (), (), (), (), ()), SimulationTypes.PowerFlow_run),
           (TimeSeriesResults(0, 0, 0, 0, (), (), (), (), (), ()), SimulationTypes.TimeSeries_run),
           (ShortCircuitResults(0, 0, 0, (), (), (), ()), SimulationTypes.ShortCircuit_run),
           (StochasticPowerFlowResults(0, 0, 0, (), (), ()), SimulationTypes.StochasticPowerFlow)
           ]

    return {tpe.value: (elm, tpe) for elm, tpe in lst}


class GcThread(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()

    def __init__(self, driver: DriverTemplate):
        QThread.__init__(self)

        # assign the driver and set the driver's reporting functions
        self.driver = driver
        self.driver.progress_signal = self.progress_signal
        self.driver.progress_text = self.progress_text
        self.driver.done_signal = self.done_signal
        self.tpe = driver.tpe

        self.results = None

        self.elapsed = 0

        self.logger = Logger()

        self.__cancel__ = False

    def get_steps(self):
        return list()

    def run(self):
        self.progress_signal.emit(0.0)

        self.driver.run()

        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        """
        Cancel the simulation
        """
        self.__cancel__ = True
        self.terminate()
        self.quit()
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled!')
        self.done_signal.emit()


class SimulationSession:

    def __init__(self, name: str = 'Session', idtag: str = None):
        """
        Constructor
        :param name: Session name
        :param idtag: Unique identifier
        """
        self.idtag: str = uuid4().hex if idtag is None else idtag

        # name of the session
        self.name: str = name

        # dictionary of drivers
        self.drivers = dict()
        self.threads: Dict[GcThread] = dict()

    def __str__(self):
        return self.name

    def clear(self):
        """
        Delete all the drivers
        """
        self.drivers = dict()

    def register(self, driver):
        """
        Register driver
        :param driver: driver to register (must have a tpe variable in it)
        """
        # register
        self.drivers[driver.tpe] = driver

    def run(self, driver, post_func=None, prog_func=None, text_func=None):
        """
        Register driver
        :param driver: driver to register (must have a tpe variable in it)
        :param post_func: Function to run after it is done
        :param prog_func: Function to display the progress
        :param text_func: Function to display text
        """

        thr = GcThread(driver)
        thr.progress_signal.connect(prog_func)
        thr.progress_text.connect(text_func)
        thr.done_signal.connect(post_func)

        # check and kill
        if driver.tpe in self.drivers.keys():
            del self.drivers[driver.tpe]
            if self.threads[driver.tpe].isRunning():
                self.threads[driver.tpe].terminate()
            del self.threads[driver.tpe]

        # register
        self.drivers[driver.tpe] = driver
        self.threads[driver.tpe] = thr

        # run!
        thr.start()

    def get_available_drivers(self):
        """
        Get a list of the available driver objects
        :return: List[Driver]
        """
        return [drv for driver_type, drv in self.drivers.items() if drv is not None]

    def exists(self, driver_type: SimulationTypes):
        """
        Get the results of the driver
        :param driver_type: driver type to look for
        :return: True / False
        """
        return driver_type in self.drivers.keys()

    def get_driver_results(self, driver_type: SimulationTypes):
        """
        Get the results of the driver
        :param driver_type: driver type
        :return: driver, results (None, None if not found)
        """
        if driver_type in self.drivers.keys():
            drv = self.drivers[driver_type]
            if hasattr(drv, 'results'):
                return drv, drv.results
            else:
                return drv, None
        else:
            return None, None

    def delete_driver(self, driver_type: SimulationTypes):
        """
        Get the results of the driver
        :param driver_type: driver type to delete
        """
        if driver_type in self.drivers.keys():
            del self.drivers[driver_type]

    def delete_driver_by_name(self, study_name: str):
        """
        Delete the driver by it's name
        :param study_name: driver name
        """
        for driver_type, drv in self.drivers.items():
            if study_name == drv.name:
                del self.drivers[driver_type]
                return

    def get_driver_by_name(self, study_name: str):
        """
        Get the driver by it's name
        :param study_name: driver name
        """
        for driver_type, drv in self.drivers.items():
            if study_name == drv.name:
                return self.drivers[driver_type]
        return None

    def get_results_model_by_name(self, study_name, study_type):
        """
        Get the results model given the study name and study type
        :param study_name: name of the study
        :param study_type: name of the study type
        :return: ResultsModel instance or None if not found
        """
        for driver_type, drv in self.drivers.items():
            if study_name == drv.name:
                if drv.results is not None:
                    return ResultsModel(drv.results.mdl(result_type=study_type))
                else:
                    print('There seem to be no results :(')
                    return None

        return None

    def register_driver_from_disk_data(self, grid, study_name: str, data_dict: dict):
        """
        Create driver with the results
        :param grid: MultiCircuit instance
        :param study_name: name of the study (i.e. Power Flow)
        :param data_dict: dictionary of data coming from the file
        :return:
        """

        # get the results' object dictionary
        drivers_dict = get_results_object_dictionary()

        if study_name in drivers_dict.keys():

            # declare a dummy driver
            drv = DriverTemplate(grid=grid)

            # set the empty results driver
            drv.results, drv.tpe = drivers_dict[study_name]
            drv.name = drv.tpe.value

            # fill in the variables
            for arr_name, arr in data_dict.items():
                setattr(drv.results, arr_name, arr)

            # perform whatever operations are needed after loading
            drv.results.consolidate_after_loading()

            self.register(drv)

    def is_this_running(self, sim_tpe: SimulationTypes):
        """
        Check if a simulation type is running
        :param sim_tpe:
        :return:
        """
        for driver_type, drv in self.threads.items():
            if drv is not None:
                if drv.isRunning():
                    if driver_type == sim_tpe:
                        return True
        return False

    def is_anything_running(self):
        """
        Check if anything is running
        :return:
        """

        for driver_type, drv in self.threads.items():
            if drv is not None:
                if drv.isRunning():
                    return True
        return False

    @property
    def power_flow(self):
        return self.get_driver_results(SimulationTypes.PowerFlow_run)[1]

    @property
    def power_flow_ts(self):
        return self.get_driver_results(SimulationTypes.TimeSeries_run)[1]

    @property
    def power_flow_cluster_ts(self):
        return self.get_driver_results(SimulationTypes.ClusteringTimeSeries_run)[1]

    @property
    def optimal_power_flow(self):
        return self.get_driver_results(SimulationTypes.OPF_run)[1]

    @property
    def optimal_power_flow_ts(self):
        return self.get_driver_results(SimulationTypes.OPFTimeSeries_run)[1]

    @property
    def short_circuit(self):
        return self.get_driver_results(SimulationTypes.ShortCircuit_run)[1]

    @property
    def linear_power_flow(self):
        return self.get_driver_results(SimulationTypes.LinearAnalysis_run)[1]

    @property
    def linear_power_flow_ts(self):
        return self.get_driver_results(SimulationTypes.LinearAnalysis_TS_run)[1]

    @property
    def contingency(self):
        return self.get_driver_results(SimulationTypes.ContingencyAnalysis_run)[1]

    @property
    def contingency_ts(self):
        return self.get_driver_results(SimulationTypes.ContingencyAnalysisTS_run)[1]

    @property
    def continuation_power_flow(self):
        return self.get_driver_results(SimulationTypes.ContinuationPowerFlow_run)[1]

    @property
    def net_transfer_capacity(self):
        return self.get_driver_results(SimulationTypes.NetTransferCapacity_run)[1]

    @property
    def net_transfer_capacity_ts(self):
        return self.get_driver_results(SimulationTypes.NetTransferCapacityTS_runy_run)[1]

    @property
    def optimal_net_transfer_capacity(self):
        return self.get_driver_results(SimulationTypes.OPF_NTC_run)[1]

    @property
    def stochastic_power_flow(self):
        return self.get_driver_results(SimulationTypes.StochasticPowerFlow)[1]

    @property
    def sigma_analysis(self):
        return self.get_driver_results(SimulationTypes.SigmaAnalysis_run)[1]

    @property
    def cascade(self):
        return self.get_driver_results(SimulationTypes.Cascade_run)[1]



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from typing import List

import numpy as np
from GridCal.Engine.Simulations.driver_types import SimulationTypes
from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
import GridCal.Engine.basic_structures as bs


class DummySignal:

    def __init__(self, tpe=str):
        self.tpe = tpe

    def emit(self, val=''):
        pass


class DriverTemplate:

    tpe = SimulationTypes.TemplateDriver
    name = 'Template'

    def __init__(self, grid: MultiCircuit, engine: bs.EngineType = bs.EngineType.GridCal):
        """

        :param grid:
        :param engine:
        """
        self.progress_signal = DummySignal()
        self.progress_text = DummySignal(str)
        self.done_signal = DummySignal()

        self.grid = grid

        self.results = None

        self.engine = engine

        self.elapsed = 0

        self.logger = Logger()

        self.__cancel__ = False

    def get_steps(self):
        return list()

    def run(self):
        pass

    def cancel(self):
        """
        Cancel the simulation
        """
        self.__cancel__ = True
        self.progress_signal.emit(0)
        self.progress_text.emit('Cancelled!')
        self.done_signal.emit()


class TimeSeriesDriverTemplate(DriverTemplate):

    def __init__(self, grid: MultiCircuit, start_=0, end_=None):

        DriverTemplate.__init__(self, grid=grid)

        self.start_ = start_

        self.indices = self.grid.time_profile

        if end_ is not None:
            self.end_ = end_
        else:
            self.end_ = len(self.grid.time_profile)

    def get_time_indices(self):
        """
        Get an array of indices of the time steps selected within the start-end interval
        :return: np.array[int]
        """
        if self.end_ is None:
            self.end_ = len(self.grid.time_profile)

        return np.arange(self.start_, self.end_ + 1)



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
This file implements a DC-OPF for time series
That means that solves the OPF problem for a complete time series at once
"""
import numpy as np
from typing import List, Tuple, Dict
import GridCal.ThirdParty.pulp as pl
from GridCal.Engine.basic_structures import ZonalGrouping
from GridCal.Engine.Core.snapshot_opf_data import SnapshotOpfData
from GridCal.Engine.Simulations.OPF.opf_templates import Opf, MIPSolvers, Logger, LpVariable
from GridCal.Engine.Devices.enumerations import TransformerControlType, ConverterControlType, HvdcControlType, GenerationNtcFormulation


def add_objective_function(Pg, Pb, LSlack, FSlack1, FSlack2, FCSlack1, FCSlack2,
                           flow_from_a1_to_a2, sum_gen_area_1, sum_gen_area_2,
                           cost_g, cost_b, cost_l, cost_br):
    """
    Add the objective function to the problem
    :param Pg: generator LpVars (ng, nt)
    :param Pb: batteries LpVars (nb, nt)
    :param LSlack: Load slack LpVars (nl, nt)
    :param FSlack1: Branch overload slack1 (m, nt)
    :param FSlack2: Branch overload slack2 (m, nt)
    :param FCSlack1: Branch contingency overload slack1 [list]
    :param FCSlack2: Branch contingency overload slack2 [List]
    :param hvdc_overload1: HVDC overload (nhvdc, nt)
    :param hvdc_overload2: HVDC overload (nhvdc, nt)
    :param hvdc_control1_slacks: HVDC control slack 1 (nhvdc, nt)
    :param hvdc_control2_slacks: HVDC control slack 2 (nhvdc, nt)
    :param cost_g: Cost of the generators (ng, nt)
    :param cost_b: Cost of the batteries (nb, nt)
    :param cost_l: Cost of the loss of load (nl, nt)
    :param cost_br: Cost of the overload (m, nt)
    :return: Nothing, just assign the objective function
    """

    f_obj = pl.lpSum(cost_g * Pg)

    f_obj += pl.lpSum(cost_b * Pb)

    f_obj += pl.lpSum(cost_l * LSlack)

    f_obj += pl.lpSum(cost_br * (FSlack1 + FSlack2))

    f_obj += pl.lpSum(FCSlack1 + FCSlack2)

    if flow_from_a1_to_a2 is not None:
        f_obj -= flow_from_a1_to_a2  # maximize

    if sum_gen_area_1 is not None:
        f_obj -= sum_gen_area_1  # maximize
        f_obj += sum_gen_area_2  # minimize

    return f_obj


def set_fix_generation(problem, Pg, P_fix, enabled_for_dispatch):
    """
    Set the generation fixed at the non dispatchable generators
    :param problem: LP problem instance
    :param Pg: Array of generation variables
    :param P_fix: Array of fixed generation values
    :param enabled_for_dispatch: array of "enables" for dispatching generators
    :return: Nothing
    """

    idx = np.where(enabled_for_dispatch == False)[0]

    pl.lpAddRestrictions2(problem=problem,
                          lhs=Pg[idx],
                          rhs=P_fix[idx],
                          name='fixed_generation',
                          op='=')


def get_power_injections(C_bus_gen, Pg, C_bus_bat, Pb, C_bus_load, LSlack, Pl):
    """
    Create the power injections per bus
    :param C_bus_gen: Bus-Generators sparse connectivity matrix (n, ng)
    :param Pg: generator LpVars (ng, nt)
    :param C_bus_bat: Bus-Batteries sparse connectivity matrix (n, nb)
    :param Pb: Batteries LpVars (nb, nt)
    :param C_bus_load: Bus-Load sparse connectivity matrix (n, nl)
    :param LSlack: Load slack LpVars (nl, nt)
    :param Pl: Load values (nl, nt)
    :return: Power injection at the buses (n, nt)
    """

    return pl.lpDot(C_bus_gen, Pg) + pl.lpDot(C_bus_bat, Pb) - pl.lpDot(C_bus_load, Pl - LSlack)


def formulate_dc_nodal_power_balance(numerical_circuit: SnapshotOpfData, problem: pl.LpProblem, theta, P):
    """
    Add the nodal power balance
    :param numerical_circuit: NumericalCircuit instance
    :param problem: LpProblem instance
    :param theta: Voltage angles LpVars (n, nt)
    :param P: Power injection at the buses LpVars (n, nt)
    :return: Nothing, the restrictions are added to the problem
    """

    # do the topological computation
    calculation_inputs = numerical_circuit.split_into_islands()

    nodal_restrictions = np.empty(numerical_circuit.nbus, dtype=object)

    # simulate each island and merge the results
    for i, calc_inpt in enumerate(calculation_inputs):

        # if there is a slack it means that there is at least one generator,
        # otherwise these equations do not make sense
        if len(calc_inpt.vd) > 0:

            # find the original indices
            bus_original_idx = np.array(calc_inpt.original_bus_idx)

            # re-pack the variables for the island and time interval
            P_island = P[bus_original_idx]  # the sizes already reflect the correct time span
            theta_island = theta[bus_original_idx]  # the sizes already reflect the correct time span
            B_island = calc_inpt.Ybus.imag

            pqpv = calc_inpt.pqpv
            vd = calc_inpt.vd

            # Add nodal power balance for the non slack nodes
            idx = bus_original_idx[pqpv]
            nodal_restrictions[idx] = pl.lpAddRestrictions2(problem=problem,
                                                            lhs=pl.lpDot(B_island[np.ix_(pqpv, pqpv)], theta_island[pqpv]),
                                                            rhs=P_island[pqpv],
                                                            name='Nodal_power_balance_pqpv_is' + str(i),
                                                            op='=')

            # Add nodal power balance for the slack nodes
            idx = bus_original_idx[vd]
            nodal_restrictions[idx] = pl.lpAddRestrictions2(problem=problem,
                                                            lhs=pl.lpDot(B_island[vd, :], theta_island),
                                                            rhs=P_island[vd],
                                                            name='Nodal_power_balance_vd_is' + str(i),
                                                            op='=')

            # slack angles equal to zero
            pl.lpAddRestrictions2(problem=problem,
                                  lhs=theta_island[vd],
                                  rhs=np.zeros(len(vd)),
                                  name='Theta_vd_zero_is' + str(i),
                                  op='=')

    return nodal_restrictions


def formulate_branch_loading_restriction(problem: pl.LpProblem, nc: SnapshotOpfData,
                                         F, T, theta, active, monitored,
                                         ratings, ratings_slack_from, ratings_slack_to):
    """
    Add the branch loading restrictions
    :param problem: LpProblem instance
    :param nc: SnapshotOpfData instance
    :param F: Array with the "from" branch indices (m)
    :param T: Array with the "to" branch indices (m)
    :param theta: voltage angles (n)
    :param active: Array of branch active states (m)
    :param monitored: Array of branch monitoring (m)
    :param ratings: Array of branch ratings (m)
    :param ratings_slack_from: Array of branch loading slack variables in the from-to sense
    :param ratings_slack_to: Array of branch loading slack variables in the to-from sense
    :return: Pbr_f: Array of power flowing through the branch (m)
             tau: Array of tap angles in the phase shifters (m)
             Pinj_tau: Array of the nodal power injections due to the phase shift (n)
    """

    nbr = len(ratings)

    # from-to branch power restriction
    Pbr_f = np.zeros(nbr, dtype=object)
    tau = np.zeros(nbr, dtype=object)
    Pinj_tau = np.zeros(nc.nbus, dtype=object)

    for m in range(nbr):
        if active[m]:

            # compute the branch susceptance
            if nc.branch_data.branch_dc[m]:
                bk = -1.0 / nc.branch_data.R[m]
            else:
                bk = -1.0 / nc.branch_data.X[m]

            # compute the flow
            if nc.branch_data.control_mode[m] == TransformerControlType.Pt:
                # is a phase shifter device (like phase shifter transformer or VSC with P control)
                tau[m] = LpVariable('Tau_{}'.format(m), nc.branch_data.theta_min[m], nc.branch_data.theta_max[m])
                Pbr_f[m] = bk * (theta[F[m]] - theta[T[m]] + tau[m])

                # power injected and subtracted due to the phase shift
                Pinj_tau[F[m]] = -bk * tau[m]
                Pinj_tau[T[m]] = bk * tau[m]
            else:
                # is a regular branch
                Pbr_f[m] = bk * (theta[F[m]] - theta[T[m]])

            if monitored[m]:
                problem.add(Pbr_f[m] <= ratings[m] + ratings_slack_from[m], 'upper_rate_{0}'.format(m))
                problem.add(-ratings[m] - ratings_slack_to[m] <= Pbr_f[m], 'lower_rate_{0}'.format(m))
        else:
            Pbr_f[m] = 0

    return Pbr_f, tau, Pinj_tau


def formulate_contingency(problem: pl.LpProblem, numerical_circuit: SnapshotOpfData, flow_f, ratings, LODF, monitor, lodf_tolerance):
    """
    Formulate contingencies
    :param problem:
    :param numerical_circuit:
    :param flow_f:
    :param ratings:
    :param LODF:
    :param monitor:
    :return:
    """
    nbr = ratings.shape[0]

    # get the indices of the branches marked for contingency
    con_br_idx = numerical_circuit.branch_data.get_contingency_enabled_indices()

    # formulate contingency flows
    # this is done in a separated loop because all te flow variables must exist beforehand
    flow_lst = list()
    indices = list()  # (t, m, contingency_m)
    overload1_lst = list()
    overload2_lst = list()

    for m in range(nbr):  # for every branch

        if monitor[m]:  # the monitor variable is pre-computed in the previous loop
            _f = numerical_circuit.branch_data.F[m]
            _t = numerical_circuit.branch_data.T[m]

            for ic, c in enumerate(con_br_idx):  # for every contingency

                if m != c and abs(LODF[m, c]) >= lodf_tolerance:

                    # compute the N-1 flow
                    contingency_flow = flow_f[m] + LODF[m, c] * flow_f[c]

                    # rating restriction in the sense from-to
                    overload1 = LpVariable("n-1_overload1_{0}_{1}".format(m, c), 0, 99999)
                    problem.add(contingency_flow <= (ratings[m] + overload1), "n-1_ft_up_rating_{0}_{1}".format(m, c))

                    # rating restriction in the sense to-from
                    overload2 = LpVariable("n-1_overload2_{0}_{1}".format(m, c), 0, 99999)
                    problem.add((-ratings[m] - overload2) <= contingency_flow, "n-1_tf_down_rating_{0}_{1}".format(m, c))

                    # store the variables
                    flow_lst.append(contingency_flow)
                    overload1_lst.append(overload1)
                    overload2_lst.append(overload2)
                    indices.append((m, c))

    return flow_lst, overload1_lst, overload2_lst, indices


def formulate_hvdc_flow(problem: pl.LpProblem, nc: SnapshotOpfData, angles, Pinj, t=0,
                        logger: Logger = Logger(), inf=999999):
    """

    :param problem:
    :param nc:
    :param angles:
    :param Pinj:
    :param t:
    :param logger:
    :param inf:
    :return:
    """
    rates = nc.hvdc_data.rate[:, t] / nc.Sbase
    F = nc.hvdc_data.get_bus_indices_f()
    T = nc.hvdc_data.get_bus_indices_t()

    flow_f = np.zeros(nc.nhvdc, dtype=object)

    angle_droop = nc.hvdc_data.get_angle_droop_in_pu_rad(nc.Sbase)[:, t]

    for i in range(nc.nhvdc):

        if nc.hvdc_data.active[i, t]:

            _f = F[i]
            _t = T[i]

            if nc.hvdc_data.control_mode[i] == HvdcControlType.type_0_free:

                if rates[i] <= 0:
                    logger.add_error('Rate = 0', 'HVDC:{0}'.format(i), rates[i])

                # formulate the hvdc flow as an AC line equivalent
                flow_f[i] = LpVariable('flow_hvdc1_' + str(i), -rates[i], rates[i])
                P0 = nc.hvdc_data.Pset[i, t] / nc.Sbase
                problem.add(flow_f[i] == P0 + angle_droop[i] * (angles[_f] - angles[_t]), 'flow_hvdc1_eq_' + str(i))
                # add the injections matching the flow
                Pinj[_f] -= flow_f[i]
                Pinj[_t] += flow_f[i]

                # rating restriction in the sense from-to: eq.17
                # overload1[i] = LpVariable('overload_hvdc1_' + str(i), 0, inf)
                # problem.add(flow_f[i] <= (rates[i] + overload1[i]), "hvdc_ft_rating_" + str(i))

                # rating restriction in the sense to-from: eq.18
                # overload2[i] = LpVariable('overload_hvdc2_' + str(i), 0, inf)
                # problem.add((-rates[i] - overload2[i]) <= flow_f[i], "hvdc_tf_rating_" + str(i))

            elif nc.hvdc_data.control_mode[i] == HvdcControlType.type_1_Pset and not nc.hvdc_data.dispatchable[i]:
                # simple injections model: The power is set by the user
                P0 = nc.hvdc_data.Pset[i, t] / nc.Sbase
                flow_f[i] = P0
                Pinj[_f] -= flow_f[i]
                Pinj[_t] += flow_f[i]

            elif nc.hvdc_data.control_mode[i] == HvdcControlType.type_1_Pset and nc.hvdc_data.dispatchable[i]:
                # simple injections model, the power is a variable and it is optimized
                P0 = LpVariable('hvdc_pf_' + str(i), -rates[i], rates[i])
                flow_f[i] = P0
                Pinj[_f] -= P0
                Pinj[_t] += P0

    return flow_f


def formulate_inter_area_flow(numerical_circuit: SnapshotOpfData, buses_areas_1, buses_areas_2, flow_f, hvdc_flow_f):
    """
    Formulate the flow that goes through the links from the area 1 (from) to the area 2 (to)
    :param numerical_circuit: SnapshotOpfData instance
    :param buses_areas_1: array of bus indices that compose the area 1 (from)
    :param buses_areas_2: array of bus indices that compose the area 2 (to)
    :param flow_f: array of branch flows
    :param hvdc_flow_f: array of HVDC links flows
    :return: Sum of flows 1->2 with the correct sign as a PuLP equation
    """
    inter_area_branches = numerical_circuit.get_inter_areas_branches(buses_areas_1=buses_areas_1,
                                                                     buses_areas_2=buses_areas_2)
    flows_ft = np.zeros(len(inter_area_branches), dtype=object)
    for i, (k, sign) in enumerate(inter_area_branches):
        flows_ft[i] = sign * flow_f[k]

    inter_area_hvdc = numerical_circuit.get_inter_areas_hvdc(buses_areas_1=buses_areas_1,
                                                             buses_areas_2=buses_areas_2)
    flows_hvdc_ft = np.zeros(len(inter_area_hvdc), dtype=object)
    for i, (k, sign) in enumerate(inter_area_hvdc):
        flows_hvdc_ft[i] = sign * hvdc_flow_f[k]

    flow_from_a1_to_a2 = pl.lpSum(flows_ft) + pl.lpSum(flows_hvdc_ft)

    return flow_from_a1_to_a2


def formulate_area_generation_summations(numerical_circuit: SnapshotOpfData, buses_areas_1, buses_areas_2, Pg):
    """
    Compute the summation of the generation in the area 1 and area 2
    :param numerical_circuit: SnapshotOpfData instance
    :param buses_areas_1: array of bus indices that compose the area 1 (from)
    :param buses_areas_2: array of bus indices that compose the area 2 (to)
    :param Pg: Array of generator variables
    :return: Summation of generation in the area 1, summation of the generation in the are 2
    """
    gens_in_a1, gens_in_a2, gens_out = numerical_circuit.get_generators_per_areas(buses_in_a1=buses_areas_1,
                                                                                  buses_in_a2=buses_areas_2)

    bat_in_a1, bat_in_a2, bat_out = numerical_circuit.get_batteries_per_areas(buses_in_a1=buses_areas_1,
                                                                              buses_in_a2=buses_areas_2)

    sum_a1 = 0
    sum_a2 = 0

    for bus_idx, gen_idx in gens_in_a1:
        sum_a1 += Pg[gen_idx]

    for bus_idx, gen_idx in gens_in_a2:
        sum_a2 += Pg[gen_idx]

    for bus_idx, gen_idx in bat_in_a1:
        sum_a1 += Pg[gen_idx]

    for bus_idx, gen_idx in bat_in_a2:
        sum_a2 += Pg[gen_idx]

    return sum_a1, sum_a2


class OpfDc(Opf):

    def __init__(self, numerical_circuit, solver_type: MIPSolvers = MIPSolvers.CBC,
                 zonal_grouping: ZonalGrouping = ZonalGrouping.NoGrouping,
                 skip_generation_limits=False, consider_contingencies=False, LODF=None,
                 lodf_tolerance=0.001, maximize_inter_area_flow=False,
                 buses_areas_1=None, buses_areas_2=None):
        """
        DC time series linear optimal power flow
        :param numerical_circuit: NumericalCircuit instance
        :param solver_type:
        :param zonal_grouping:
        :param skip_generation_limits:
        :param consider_contingencies:
        :param LODF:
        :param maximize_inter_area_flow:
        :param buses_areas_1:
        :param buses_areas_2:
        """
        Opf.__init__(self, numerical_circuit=numerical_circuit, solver_type=solver_type)
        self.zonal_grouping = zonal_grouping
        self.skip_generation_limits = skip_generation_limits
        self.consider_contingencies = consider_contingencies
        self.LODF = LODF
        self.lodf_tolerance = lodf_tolerance

        self.maximize_inter_area_flow = maximize_inter_area_flow
        self.buses_areas_1: List[int] = buses_areas_1
        self.buses_areas_2: List[int] = buses_areas_2

    def formulate(self):
        """
        Formulate the AC OPF time series in the non-sequential fashion (all to the solver_type at once)
        :return: PuLP Problem instance
        """
        # --------------------------------------------------------------------------------------------------------------
        # general indices
        n = self.numerical_circuit.nbus
        m = self.numerical_circuit.nbr
        ng = self.numerical_circuit.ngen
        nb = self.numerical_circuit.nbatt
        nl = self.numerical_circuit.nload
        Sbase = self.numerical_circuit.Sbase

        # battery
        Pb_max = self.numerical_circuit.battery_pmax / Sbase
        Pb_min = self.numerical_circuit.battery_pmin / Sbase
        cost_b = self.numerical_circuit.battery_cost

        # generator
        if self.skip_generation_limits:
            Pg_max = np.zeros(self.numerical_circuit.ngen) + 99999
            Pg_min = np.zeros(self.numerical_circuit.ngen) - 99999
        else:
            Pg_max = self.numerical_circuit.generator_pmax / Sbase
            Pg_min = self.numerical_circuit.generator_pmin / Sbase
        cost_g = self.numerical_circuit.generator_cost
        P_fix = self.numerical_circuit.generator_p / Sbase
        enabled_for_dispatch = self.numerical_circuit.generator_dispatchable

        # load
        Pl = (self.numerical_circuit.load_active * self.numerical_circuit.load_s.real) / Sbase
        cost_l = self.numerical_circuit.load_cost

        # branch
        branch_active = self.numerical_circuit.branch_data.branch_active[:, 0]
        branch_monitored = self.numerical_circuit.branch_data.monitor_loading
        branch_ratings = self.numerical_circuit.branch_rates / Sbase

        cost_br = self.numerical_circuit.branch_cost

        # create LP variables ------------------------------------------------------------------------------------------
        Pg = pl.lpMakeVars(name='Pg', shape=ng, lower=Pg_min, upper=Pg_max)
        Pb = pl.lpMakeVars(name='Pb', shape=nb, lower=Pb_min, upper=Pb_max)
        load_slack = pl.lpMakeVars(name='LSlack', shape=nl, lower=0, upper=None)
        theta = pl.lpMakeVars(name='theta', shape=n,
                              lower=self.numerical_circuit.bus_data.angle_min,
                              upper=self.numerical_circuit.bus_data.angle_max)
        branch_rating_slack1 = pl.lpMakeVars(name='FSlack1', shape=m, lower=0, upper=None)
        branch_rating_slack2 = pl.lpMakeVars(name='FSlack2', shape=m, lower=0, upper=None)

        # declare problem ----------------------------------------------------------------------------------------------
        self.problem = pl.LpProblem(name='DC_OPF')

        # set the fixed generation values ------------------------------------------------------------------------------
        set_fix_generation(problem=self.problem, Pg=Pg, P_fix=P_fix, enabled_for_dispatch=enabled_for_dispatch)

        # compute the nodal power injections ---------------------------------------------------------------------------
        P = get_power_injections(C_bus_gen=self.numerical_circuit.generator_data.C_bus_gen, Pg=Pg,
                                 C_bus_bat=self.numerical_circuit.battery_data.C_bus_batt, Pb=Pb,
                                 C_bus_load=self.numerical_circuit.load_data.C_bus_load,
                                 LSlack=load_slack, Pl=Pl)

        # formulate the simple HVDC models -----------------------------------------------------------------------------
        hvdc_flow_f = formulate_hvdc_flow(problem=self.problem,
                                          nc=self.numerical_circuit,
                                          angles=theta,
                                          Pinj=P,
                                          t=0,
                                          logger=self.logger,
                                          inf=999999)

        # add the branch loading restriction ---------------------------------------------------------------------------
        flow_f, tau, Pinj_tau = formulate_branch_loading_restriction(problem=self.problem,
                                                                     nc=self.numerical_circuit,
                                                                     F=self.numerical_circuit.F,
                                                                     T=self.numerical_circuit.T,
                                                                     theta=theta,
                                                                     active=branch_active,
                                                                     monitored=branch_monitored,
                                                                     ratings=branch_ratings,
                                                                     ratings_slack_from=branch_rating_slack1,
                                                                     ratings_slack_to=branch_rating_slack2)

        # add the DC grid restrictions (with real slack losses) --------------------------------------------------------
        self.nodal_restrictions = formulate_dc_nodal_power_balance(numerical_circuit=self.numerical_circuit,
                                                                   problem=self.problem,
                                                                   theta=theta,
                                                                   P=P + Pinj_tau)  # add the phase shift injections

        # formulate contingencies --------------------------------------------------------------------------------------
        if self.consider_contingencies:
            con_flow_lst, con_overload1_lst, con_overload2_lst, \
            con_idx = formulate_contingency(problem=self.problem,
                                            numerical_circuit=self.numerical_circuit,
                                            flow_f=flow_f,
                                            ratings=branch_ratings,
                                            LODF=self.LODF,
                                            monitor=self.numerical_circuit.branch_data.monitor_loading,
                                            lodf_tolerance=self.lodf_tolerance)
        else:
            con_flow_lst = list()
            con_idx = list()
            con_overload1_lst = list()
            con_overload2_lst = list()

        # maximize the power from->to ----------------------------------------------------------------------------------
        if self.maximize_inter_area_flow:
            flow_from_a1_to_a2 = formulate_inter_area_flow(numerical_circuit=self.numerical_circuit,
                                                           buses_areas_1=self.buses_areas_1,
                                                           buses_areas_2=self.buses_areas_2,
                                                           flow_f=flow_f,
                                                           hvdc_flow_f=hvdc_flow_f)

            sum_gen_area_1, sum_gen_area_2 = formulate_area_generation_summations(numerical_circuit=self.numerical_circuit,
                                                                                  buses_areas_1=self.buses_areas_1,
                                                                                  buses_areas_2=self.buses_areas_2,
                                                                                  Pg=Pg)

        else:
            flow_from_a1_to_a2 = None
            sum_gen_area_1 = None
            sum_gen_area_2 = None

        # add the objective function -----------------------------------------------------------------------------------
        self.problem += add_objective_function(Pg=Pg,
                                               Pb=Pb,
                                               LSlack=load_slack,
                                               FSlack1=branch_rating_slack1,
                                               FSlack2=branch_rating_slack2,
                                               FCSlack1=con_overload1_lst,
                                               FCSlack2=con_overload2_lst,
                                               flow_from_a1_to_a2=flow_from_a1_to_a2,
                                               sum_gen_area_1=sum_gen_area_1,
                                               sum_gen_area_2=sum_gen_area_2,
                                               cost_g=cost_g,
                                               cost_b=cost_b,
                                               cost_l=cost_l,
                                               cost_br=cost_br)

        # Assign variables to keep
        # transpose them to be in the format of GridCal: time, device
        self.theta = theta
        self.Pg = Pg
        self.Pb = Pb
        self.Pl = Pl

        self.Pinj = P

        self.phase_shift = tau

        self.hvdc_flow = hvdc_flow_f

        self.load_shedding = load_slack
        self.s_from = flow_f
        self.s_to = -flow_f
        self.overloads = branch_rating_slack1 + branch_rating_slack2
        self.rating = branch_ratings

        self.contingency_flows_list = con_flow_lst
        self.contingency_indices_list = con_idx  # [(t, m, c), ...]
        self.contingency_flows_slacks_list = con_overload1_lst


if __name__ == '__main__':
    from GridCal.Engine.basic_structures import BranchImpedanceMode
    from GridCal.Engine.IO.file_handler import FileOpen
    from GridCal.Engine.Core.snapshot_opf_data import compile_snapshot_opf_circuit

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/grid_2_islands.xlsx'

    main_circuit = FileOpen(fname).open()

    # main_circuit.buses[3].controlled_generators[0].enabled_dispatch = False

    numerical_circuit_ = compile_snapshot_opf_circuit(circuit=main_circuit,
                                                      apply_temperature=False,
                                                      branch_tolerance_mode=BranchImpedanceMode.Specified)

    problem = OpfDc(numerical_circuit=numerical_circuit_)

    print('Solving...')
    status = problem.solve()

    # print("Status:", status)

    v = problem.get_voltage()
    print('Angles\n', np.angle(v))

    l = problem.get_loading()
    print('Branch loading\n', l)

    g = problem.get_generator_power()
    print('Gen power\n', g)

    pr = problem.get_shadow_prices()
    print('Nodal prices \n', pr)



# PuLP : Python LP Modeler
# Version 1.4.2

# Copyright (c) 2002-2005, Jean-Sebastien Roy (js@jeannot.org)
# Modifications Copyright (c) 2007- Stuart Anthony Mitchell (s.mitchell@auckland.ac.nz)
# $Id:solvers.py 1791 2008-04-23 22:54:34Z smit023 $

# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE."""

from GridCal.ThirdParty.pulp.solvers import *

# get the glpk name in global scope
glpk = None



class GLPK_CMD(LpSolver_CMD):
    """The GLPK LP solver_type"""
    def defaultPath(self):
        return self.executableExtension(glpk_path)

    def available(self):
        """True if the solver_type is available"""
        return self.executable(self.path)

    def actualSolve(self, lp):
        """Solve a well formulated lp problem"""
        if not self.executable(self.path):
            raise PulpSolverError("PuLP: cannot execute "+self.path)
        if not self.keepFiles:
            uuid = uuid4().hex
            tmpLp = os.path.join(self.tmpDir, "%s-pulp.lp" % uuid)
            tmpSol = os.path.join(self.tmpDir, "%s-pulp.sol" % uuid)
        else:
            tmpLp = lp.name+"-pulp.lp"
            tmpSol = lp.name+"-pulp.sol"
        lp.writeLP(tmpLp, writeSOS = 0)
        proc = ["glpsol", "--cpxlp", tmpLp, "-o", tmpSol]
        if not self.mip: proc.append('--nomip')
        proc.extend(self.options)

        self.solution_time = clock()
        if not self.msg:
            proc[0] = self.path
            pipe = open(os.devnull, 'w')
            if operating_system == 'win':
                # Prevent flashing windows if used from a GUI application
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                rc = subprocess.call(proc, stdout = pipe, stderr = pipe,
                                     startupinfo = startupinfo)
            else:
                rc = subprocess.call(proc, stdout = pipe, stderr = pipe)
            if rc:
                raise PulpSolverError("PuLP: Error while trying to execute "+self.path)
            pipe.close()
        else:
            if os.name != 'nt':
                rc = os.spawnvp(os.P_WAIT, self.path, proc)
            else:
                rc = os.spawnv(os.P_WAIT, self.executable(self.path), proc)
            if rc == 127:
                raise PulpSolverError("PuLP: Error while trying to execute "+self.path)
        self.solution_time += clock()

        if not os.path.exists(tmpSol):
            raise PulpSolverError("PuLP: Error while executing "+self.path)
        lp.status, values = self.readsol(tmpSol)
        lp.assignVarsVals(values)
        if not self.keepFiles:
            try: os.remove(tmpLp)
            except: pass
            try: os.remove(tmpSol)
            except: pass
        return lp.status

    def readsol(self,filename):
        """Read a GLPK solution file"""
        with open(filename) as f:
            f.readline()
            rows = int(f.readline().split()[1])
            cols = int(f.readline().split()[1])
            f.readline()
            statusString = f.readline()[12:-1]
            glpkStatus = {
                "INTEGER OPTIMAL":LpStatusOptimal,
                "INTEGER NON-OPTIMAL":LpStatusOptimal,
                "OPTIMAL":LpStatusOptimal,
                "INFEASIBLE (FINAL)":LpStatusInfeasible,
                "INTEGER UNDEFINED":LpStatusUndefined,
                "UNBOUNDED":LpStatusUnbounded,
                "UNDEFINED":LpStatusUndefined,
                "INTEGER EMPTY":LpStatusInfeasible
                }
            #print "statusString ",statusString
            if statusString not in glpkStatus:
                raise PulpSolverError("Unknown status returned by GLPK")
            status = glpkStatus[statusString]
            isInteger = statusString in ["INTEGER NON-OPTIMAL","INTEGER OPTIMAL","INTEGER UNDEFINED"]
            values = {}
            for i in range(4): f.readline()
            for i in range(rows):
                line = f.readline().split()
                if len(line) ==2: f.readline()
            for i in range(3):
                f.readline()
            for i in range(cols):
                line = f.readline().split()
                name = line[1]
                if len(line) ==2: line = [0,0]+f.readline().split()
                if isInteger:
                    if line[2] == "*": value = int(float(line[3]))
                    else: value = float(line[2])
                else:
                    value = float(line[3])
                values[name] = value
        return status, values

GLPK = GLPK_CMD


class PYGLPK(LpSolver):
    """
    The glpk LP/MIP solver_type (via its python interface)

    Copyright Christophe-Marie Duquesne 2012

    The glpk variables are available (after a solve) in var.solverVar
    The glpk constraints are available in constraint.solverConstraint
    The Model is in prob.solverModel
    """
    try:
        #import the model into the global scope
        global glpk
        import glpk.glpkpi as glpk
    except:
        def available(self):
            """True if the solver_type is available"""
            return False
        def actualSolve(self, lp, callback = None):
            """Solve a well formulated lp problem"""
            raise PulpSolverError("GLPK: Not Available")
    else:
        def __init__(self,
                    mip = True,
                    msg = True,
                    timeLimit = None,
                    epgap = None,
                    **solverParams):
            """
            Initializes the glpk solver_type.

            @param mip: if False the solver_type will solve a MIP as an LP
            @param msg: displays information from the solver_type to stdout
            @param timeLimit: not handled
            @param epgap: not handled
            @param solverParams: not handled
            """
            LpSolver.__init__(self, mip, msg)
            if not self.msg:
                glpk.glp_term_out(glpk.GLP_OFF)

        def findSolutionValues(self, lp):
            prob = lp.solverModel
            if self.mip and self.hasMIPConstraints(lp.solverModel):
                solutionStatus = glpk.glp_mip_status(prob)
            else:
                solutionStatus = glpk.glp_get_status(prob)
            glpkLpStatus = {glpk.GLP_OPT: LpStatusOptimal,
                                   glpk.GLP_UNDEF: LpStatusUndefined,
                                   glpk.GLP_FEAS: LpStatusNotSolved,
                                   glpk.GLP_INFEAS: LpStatusInfeasible,
                                   glpk.GLP_NOFEAS: LpStatusInfeasible,
                                   glpk.GLP_UNBND: LpStatusUnbounded
                                   }
            #populate pulp solution values
            for var in lp.variables():
                if self.mip and self.hasMIPConstraints(lp.solverModel):
                    var.varValue = glpk.glp_mip_col_val(prob, var.glpk_index)
                else:
                    var.varValue = glpk.glp_get_col_prim(prob, var.glpk_index)
                var.dj = glpk.glp_get_col_dual(prob, var.glpk_index)
            #put pi and slack variables against the constraints
            for constr in lp.constraints.values():
                if self.mip and self.hasMIPConstraints(lp.solverModel):
                    row_val = glpk.glp_mip_row_val(prob, constr.glpk_index)
                else:
                    row_val = glpk.glp_get_row_prim(prob, constr.glpk_index)
                constr.slack = -constr.constant - row_val
                constr.pi = glpk.glp_get_row_dual(prob, constr.glpk_index)
            lp.resolveOK = True
            for var in lp.variables():
                var.isModified = False
            lp.status = glpkLpStatus.get(solutionStatus,
                    LpStatusUndefined)
            return lp.status

        def available(self):
            """True if the solver_type is available"""
            return True

        def hasMIPConstraints(self, solverModel):
            return (glpk.glp_get_num_int(solverModel) > 0 or
                    glpk.glp_get_num_bin(solverModel) > 0)

        def callSolver(self, lp, callback = None):
            """Solves the problem with glpk
            """
            self.solveTime = -clock()
            glpk.glp_adv_basis(lp.solverModel, 0)
            glpk.glp_simplex(lp.solverModel, None)
            if self.mip and self.hasMIPConstraints(lp.solverModel):
                status = glpk.glp_get_status(lp.solverModel)
                if status in (glpk.GLP_OPT, glpk.GLP_UNDEF, glpk.GLP_FEAS):
                    glpk.glp_intopt(lp.solverModel, None)
            self.solveTime += clock()

        def buildSolverModel(self, lp):
            """
            Takes the pulp lp model and translates it into a glpk model
            """

            log.debug("create the glpk model")
            prob = glpk.glp_create_prob()

            glpk.glp_set_prob_name(prob, lp.name)
            log.debug("set the sense of the problem")

            if lp.sense == LpMaximize:
                glpk.glp_set_obj_dir(prob, glpk.GLP_MAX)

            log.debug("add the constraints to the problem")

            glpk.glp_add_rows(prob, len(list(lp.constraints.keys())))

            for i, v in enumerate(lp.constraints.items(), start=1):
                name, constraint = v
                glpk.glp_set_row_name(prob, i, name)
                if constraint.sense == LpConstraintLE:
                    glpk.glp_set_row_bnds(prob, i, glpk.GLP_UP, 0.0, -constraint.constant)
                elif constraint.sense == LpConstraintGE:
                    glpk.glp_set_row_bnds(prob, i, glpk.GLP_LO, -constraint.constant, 0.0)
                elif constraint.sense == LpConstraintEQ:
                    glpk.glp_set_row_bnds(prob, i, glpk.GLP_FX, -constraint.constant, -constraint.constant)
                else:
                    raise PulpSolverError('Detected an invalid constraint type')
                constraint.glpk_index = i

            log.debug("add the variables to the problem")

            glpk.glp_add_cols(prob, len(lp.variables()))

            for j, var in enumerate(lp.variables(), start=1):

                glpk.glp_set_col_name(prob, j, var.name)
                lb = 0.0
                ub = 0.0
                t = glpk.GLP_FR

                if not var.lowBound is None:
                    lb = var.lowBound
                    t = glpk.GLP_LO

                if not var.upBound is None:
                    ub = var.upBound
                    t = glpk.GLP_UP

                if not var.upBound is None and not var.lowBound is None:
                    if ub == lb:
                        t = glpk.GLP_FX
                    else:
                        t = glpk.GLP_DB

                glpk.glp_set_col_bnds(prob, j, t, lb, ub)

                if var.cat == LpInteger:
                    glpk.glp_set_col_kind(prob, j, glpk.GLP_IV)
                    assert glpk.glp_get_col_kind(prob, j) == glpk.GLP_IV

                var.glpk_index = j

            log.debug("set the objective function")

            for var in lp.variables():
                value = lp.objective.get(var)
                if value:
                    glpk.glp_set_obj_coef(prob, var.glpk_index, value)

            log.debug("set the problem matrix")

            for constraint in lp.constraints.values():
                l = len(list(constraint.items()))
                ind = glpk.intArray(l + 1)
                val = glpk.doubleArray(l + 1)

                for j, v in enumerate(constraint.items(), start=1):
                    var, value = v
                    ind[j] = var.glpk_index
                    val[j] = value

                glpk.glp_set_mat_row(prob, constraint.glpk_index, l, ind, val)

            lp.solverModel = prob

        def actualSolve(self, lp, callback = None):
            """
            Solve a well formulated lp problem

            creates a glpk model, variables and constraints and attaches
            them to the lp model which it then solves
            """
            self.buildSolverModel(lp)

            # set the initial solution
            log.debug("Solve the Model using glpk")
            self.callSolver(lp, callback=callback)

            # get the solution information
            solutionStatus = self.findSolutionValues(lp)

            for var in lp.variables():
                var.modified = False

            for constraint in lp.constraints.values():
                constraint.modified = False

            return solutionStatus

        def actualResolve(self, lp, callback = None):
            """
            Solve a well formulated lp problem

            uses the old solver_type and modifies the rhs of the modified
            constraints
            """
            log.debug("Resolve the Model using glpk")
            for constraint in lp.constraints.values():
                i = constraint.glpk_index
                if constraint.modified:
                    if constraint.sense == LpConstraintLE:
                        glpk.glp_set_row_bnds(prob, i, glpk.GLP_UP, 0.0, -constraint.constant)
                    elif constraint.sense == LpConstraintGE:
                        glpk.glp_set_row_bnds(prob, i, glpk.GLP_LO, -constraint.constant, 0.0)
                    elif constraint.sense == LpConstraintEQ:
                        glpk.glp_set_row_bnds(prob, i, glpk.GLP_FX, -constraint.constant, -constraint.constant)
                    else:
                        raise PulpSolverError('Detected an invalid constraint type')
            self.callSolver(lp, callback=callback)

            # get the solution information
            solutionStatus = self.findSolutionValues(lp)

            for var in lp.variables():
                var.modified = False

            for constraint in lp.constraints.values():
                constraint.modified = False

            return solutionStatus




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import time
import scipy

from GridCal.Engine.Simulations.sparse_solve import get_sparse_type, get_linear_solver
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.common_functions import *
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults

linear_solver = get_linear_solver()
sparse = get_sparse_type()
scipy.ALLOW_THREADS = True
np.set_printoptions(precision=8, suppress=True, linewidth=320)


def Jacobian_decoupled(Ybus, V, Ibus, pq, pvpq):
    """
    Computes the decoupled Jacobian matrices
    Args:
        Ybus: Admittance matrix
        V: Array of nodal voltages
        Ibus: Array of nodal current injections
        pq: Array with the indices of the PQ buses
        pvpq: Array with the indices of the PV and PQ buses

    Returns: J11, J22
    """
    I = Ybus * V - Ibus

    diagV = sp.diags(V)
    diagI = sp.diags(I)
    diagVnorm = sp.diags(V / np.abs(V))

    dS_dVm = diagV * np.conj(Ybus * diagVnorm) + np.conj(diagI) * diagVnorm
    dS_dVa = 1.0j * diagV * np.conj(diagI - Ybus * diagV)

    J11 = dS_dVa[np.ix_(pvpq, pvpq)].real
    J22 = dS_dVm[np.ix_(pq, pq)].imag

    return J11, J22


def NRD_LS(Ybus, S0, V0, I0, Y0, pv, pq, tol, max_it=15,
           acceleration_parameter=0.5, error_registry=None, verbose=False) -> NumericPowerFlowResults:
    """
    Solves the power flow using a full Newton's method with backtrack correction.
    @Author: Santiago Peñate Vera
    :param Ybus: Admittance matrix
    :param S0: Array of nodal power injections
    :param V0: Array of nodal voltages (initial solution)
    :param I0: Array of nodal current injections
    :param pv: Array with the indices of the PV buses
    :param pq: Array with the indices of the PQ buses
    :param tol: Tolerance
    :param max_it: Maximum number of iterations
    :param acceleration_parameter: parameter used to correct the "bad" iterations, typically 0.5
    :param error_registry: list to store the error for plotting
    :return: Voltage solution, converged?, error, calculated power injections
    """

    start = time.time()

    use_norm_error = True

    # initialize
    back_track_counter = 0
    back_track_iterations = 0
    converged = 0
    iter_ = 0
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # evaluate F(x0)
    Sbus = compute_zip_power(S0, I0, Y0, Vm)
    Scalc = compute_power(Ybus, V)
    f = compute_fx(Scalc, Sbus, pvpq, pq)

    # check tolerance
    if use_norm_error:
        norm_f = np.linalg.norm(f, np.Inf)
    else:
        norm_f = 0.5 * f.dot(f)

    if error_registry is not None:
        error_registry.append(norm_f)

    if norm_f < tol:
        converged = 1

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        # evaluate Jacobian
        J1, J4 = Jacobian_decoupled(Ybus, V, I0, pq, pvpq)

        # compute update step and reassign the solution vector
        dVa[pvpq] = linear_solver(J1, f[pvpq])
        dVm[pq] = linear_solver(J4, f[pq])

        # update voltage the Newton way (mu=1)
        mu_ = 1.0
        Vm -= mu_ * dVm
        Va -= mu_ * dVa
        Vnew = polar_to_rect(Vm, Va)

        # compute the mismatch function f(x_new)
        Sbus = compute_zip_power(S0, I0, Y0, Vm)
        Scalc = compute_power(Ybus, Vnew)
        f_new = compute_fx(Scalc, Sbus, pvpq, pq)

        if use_norm_error:
            norm_f_new = np.linalg.norm(f_new, np.Inf)
        else:
            norm_f_new = 0.5 * f_new.dot(f_new)

        if error_registry is not None:
            error_registry.append(norm_f_new)

        cond = norm_f_new > norm_f  # condition to back track (no improvement at all)

        if not cond:
            back_track_counter += 1

        l_iter = 0
        while not cond and l_iter < 10 and mu_ > 0.01:
            # line search back
            # update voltage with a closer value to the last value in the Jacobian direction
            mu_ *= acceleration_parameter
            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            Vnew = polar_to_rect(Vm, Va)

            # compute the mismatch function f(x_new)
            Sbus = compute_zip_power(S0, I0, Y0, Vm)
            Scalc = compute_power(Ybus, Vnew)
            f_new = compute_fx(Scalc, Sbus, pvpq, pq)

            if use_norm_error:
                norm_f_new = np.linalg.norm(f_new, np.Inf)
            else:
                norm_f_new = 0.5 * f_new.dot(f_new)

            cond = norm_f_new > norm_f

            if error_registry is not None:
                error_registry.append(norm_f_new)

            l_iter += 1
            back_track_iterations += 1

        # update calculation variables
        V = Vnew
        f = f_new

        # check for convergence
        if use_norm_error:
            norm_f = np.linalg.norm(f_new, np.Inf)
        else:
            norm_f = 0.5 * f_new.dot(f_new)

        if error_registry is not None:
            error_registry.append(norm_f)

        if norm_f < tol:
            converged = 1

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, converged, norm_f, Scalc, None, None, None, None, None, None, iter_, elapsed)




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

from GridCal.Gui.GridEditorWidget.generic_graphics import ACTIVE, DEACTIVATED, FONT_SCALE, EMERGENCY
# from GridCal.Gui.GridEditorWidget.editor import GridEditor


class TerminalItem(QGraphicsRectItem):
    """
    Represents a connection point to a subsystem
    """

    def __init__(self, name, editor=None, parent=None, h=10.0, w=10.0):
        """

        @param name:
        @param editor:
        @param parent:
        """

        QGraphicsRectItem.__init__(self, QRectF(-6.0, -6.0, h, w), parent)
        self.setCursor(QCursor(Qt.CrossCursor))

        # Properties:
        self.color = ACTIVE['color']
        self.pen_width = 2
        self.style = ACTIVE['style']
        self.setBrush(Qt.darkGray)
        self.setPen(QPen(self.color, self.pen_width, self.style))

        # terminal parent object
        self.parent = parent

        self.hosting_connections = list()

        self.editor = editor

        # Name:
        self.name = name
        self.posCallbacks = list()
        self.setFlag(self.ItemSendsScenePositionChanges, True)

    @property
    def w(self):
        return self.rect().width()

    @property
    def h(self):
        return self.rect().height()

    @property
    def x(self):
        return self.pos().x()

    @property
    def y(self):
        return self.pos().y()

    @property
    def xc(self):
        return self.pos().x() - self.w / 2

    @property
    def yc(self):
        return self.pos().y() - self.h / 2

    def update(self):

        self.process_callbacks(self.parent.pos() + self.pos())

    def process_callbacks(self, value):
        """

        :param value:
        :return:
        """
        w = self.rect().width()
        h2 = self.rect().height() / 2.0
        n = len(self.posCallbacks)
        dx = w / (n + 1)
        for i, call_back in enumerate(self.posCallbacks):
            call_back(value + QPointF((i + 1) * dx, h2))

    def itemChange(self, change, value):
        """

        @param change:
        @param value: This is a QPointF object with the coordinates of the upper left corner of the TerminalItem
        @return:
        """
        if change == self.ItemScenePositionHasChanged:

            self.process_callbacks(value)

            return value

        else:
            return super(TerminalItem, self).itemChange(change, value)

    def mousePressEvent(self, event):
        """
        Start a connection
        Args:
            event:

        Returns:

        """
        self.editor.start_connection(self)
        if self.editor.started_branch is not None:
            self.hosting_connections.append(self.editor.started_branch)

    def remove_all_connections(self):
        """
        Removes all the terminal connections
        Returns:

        """
        n = len(self.hosting_connections)
        for i in range(n - 1, -1, -1):
            self.hosting_connections[i].remove_widget()
            self.hosting_connections.pop(i)


class HandleItem(QGraphicsEllipseItem):
    """
    A handle that can be moved by the mouse: Element to resize the boxes
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """
        QGraphicsEllipseItem.__init__(self, QRectF(-4, -4, 8, 8), parent)

        self.posChangeCallbacks = list()
        self.setBrush(Qt.red)
        self.setFlag(self.ItemIsMovable, True)
        self.setFlag(self.ItemSendsScenePositionChanges, True)
        self.setCursor(QCursor(Qt.SizeFDiagCursor))

    def itemChange(self, change, value):
        """

        @param change:
        @param value:
        @return:
        """
        if change == self.ItemPositionChange:
            x, y = value.x(), value.y()

            # This cannot be a signal because this is not a QObject
            for cb in self.posChangeCallbacks:
                res = cb(x, y)
                if res:
                    x, y = res
                    value = QPointF(x, y)
            return value

        # Call superclass method:
        return super(HandleItem, self).itemChange(change, value)




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gui.ui',
# licensing of 'gui.ui' applies.
#
# Created: Sun Mar  8 13:42:31 2020
#      by: pyside2-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(841, 518)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/Icons/icons/sigma.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        MainWindow.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter_3 = QtWidgets.QSplitter(self.centralwidget)
        self.splitter_3.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_3.setObjectName("splitter_3")
        self.frame_8 = QtWidgets.QFrame(self.splitter_3)
        self.frame_8.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_8.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_8.setObjectName("frame_8")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.frame_8)
        self.verticalLayout_7.setContentsMargins(-1, 0, -1, -1)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.tableView = QtWidgets.QTableView(self.frame_8)
        self.tableView.setObjectName("tableView")
        self.verticalLayout_7.addWidget(self.tableView)
        self.PlotFrame = QtWidgets.QFrame(self.splitter_3)
        self.PlotFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.PlotFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.PlotFrame.setObjectName("PlotFrame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.PlotFrame)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName("plotwidget")
        self.horizontalLayout.addWidget(self.plotwidget)
        self.verticalLayout.addWidget(self.splitter_3)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 841, 22))
        self.menubar.setObjectName("menubar")
        self.menuActions = QtWidgets.QMenu(self.menubar)
        self.menuActions.setObjectName("menuActions")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionCopy_to_clipboard = QtWidgets.QAction(MainWindow)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/Icons/icons/copy.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionCopy_to_clipboard.setIcon(icon1)
        self.actionCopy_to_clipboard.setObjectName("actionCopy_to_clipboard")
        self.actionSave = QtWidgets.QAction(MainWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/Icons/icons/import_profiles.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionSave.setIcon(icon2)
        self.actionSave.setObjectName("actionSave")
        self.menuActions.addAction(self.actionCopy_to_clipboard)
        self.menuActions.addAction(self.actionSave)
        self.menubar.addAction(self.menuActions.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QtWidgets.QApplication.translate("MainWindow", "MainWindow", None, -1))
        self.menuActions.setTitle(QtWidgets.QApplication.translate("MainWindow", "Actions", None, -1))
        self.actionCopy_to_clipboard.setText(QtWidgets.QApplication.translate("MainWindow", "Copy to clipboard", None, -1))
        self.actionSave.setText(QtWidgets.QApplication.translate("MainWindow", "Save", None, -1))

from .matplotlibwidget import MatplotlibWidget
from .icons_rc import *

if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec_())




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
from numpy import sqrt
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Devices.bus import Bus
from GridCal.Engine.Devices.enumerations import BranchType, TransformerControlType

from GridCal.Engine.Devices.editable_device import EditableDevice, DeviceType, GCProp


class Transformer3W(EditableDevice):

    def __init__(self, bus1: Bus = None, bus2: Bus = None, bus3: Bus = None, V1=10, V2=10, V3=10,
                 name='Branch', idtag=None, code='', active=True):
        EditableDevice.__init__(self,
                                name=name,
                                idtag=idtag,
                                active=active,
                                code=code,
                                device_type=DeviceType.Transformer3WDevice,
                                editable_headers={'name': GCProp('', str, 'Name of the branch.'),
                                                  'idtag': GCProp('', str, 'Unique ID', False),
                                                  'code': GCProp('', str, 'Secondary ID'),
                                                  'bus1': GCProp('', DeviceType.BusDevice, 'Bus 1.'),
                                                  'bus2': GCProp('', DeviceType.BusDevice, 'Bus 2.'),
                                                  'bus3': GCProp('', DeviceType.BusDevice, 'Bus 3.'),
                                                  'active': GCProp('', bool, 'Is the branch active?'),
                                                  'V1': GCProp('kV', float, 'Side 1 rating'),
                                                  'V2': GCProp('kV', float, 'Side 2 rating'),
                                                  'V3': GCProp('kV', float, 'Side 3 rating'),
                                                  },
                                non_editable_attributes=['bus1', 'bus2', 'bus3'],
                                properties_with_profile={'active': 'active_prof'})

        self.bus1 = bus1
        self.bus2 = bus2
        self.bus3 = bus3

        self.V1 = V1
        self.V2 = V2
        self.V3 = V3

        self.x = 0
        self.y = 0
        self.h = 0
        self.w = 0

    def retrieve_graphic_position(self):
        """
        Get the position set by the graphic object into this object's variables
        :return: Nothing
        """
        if self.graphic_obj is not None:
            self.x = int(self.graphic_obj.pos().x())
            self.y = int(self.graphic_obj.pos().y())



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
import numba as nb
from matplotlib import pyplot as plt

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
from GridCal.Engine.Simulations.results_table import ResultsTable
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Core.snapshot_pf_data import compile_snapshot_circuit
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.helm_power_flow import helm_coefficients_josep, sigma_function
from GridCal.Engine.Simulations.driver_template import DriverTemplate
from GridCal.Engine.Simulations.driver_types import SimulationTypes


class SigmaAnalysisResults:

    def __init__(self, n):

        self.n = n

        self.name = 'Power flow'

        self.lambda_value = 1.0

        self.bus_names = np.zeros(n, dtype=object)

        self.Sbus = np.zeros(n, dtype=complex)

        self.distances = np.zeros(n, dtype=float) + 0.25  # the default distance is 0.25

        self.sigma_re = np.zeros(n, dtype=float)

        self.sigma_im = np.zeros(n, dtype=float)

        self.available_results = [ResultTypes.SigmaReal,
                                  ResultTypes.SigmaImag,
                                  ResultTypes.SigmaDistances,
                                  ResultTypes.SigmaPlusDistances]

        self.elapsed = 0

        self.convergence_reports = list()

    def apply_from_island(self, results: "SigmaAnalysisResults", b_idx):
        """
        Apply results from another island circuit to the circuit results represented
        here.

        Arguments:

            **results**: PowerFlowResults

            **b_idx**: bus original indices

            **elm_idx**: branch original indices
        """
        self.Sbus[b_idx] = results.Sbus

        self.distances[b_idx] = results.distances

        self.sigma_re[b_idx] = results.sigma_re

        self.sigma_im[b_idx] = results.sigma_im

    def plot(self, fig, ax, n_points=1000):
        """
        Plot the sigma analysis
        :param fig: Matplotlib figure. If None, one will be created
        :param ax: Matplotlib Axis
        :param n_points: number of points in the curve
        """
        if ax is None:
            fig = plt.figure(figsize=(8, 7))
            ax = fig.add_subplot(111)

        sx = np.linspace(-0.25, np.max(self.sigma_re) + 0.1, n_points)
        sy1 = np.sqrt(0.25 + sx)
        sy2 = -np.sqrt(0.25 + sx)
        names = self.bus_names

        ax.plot(sx, sy1, 'k', linewidth=2)
        ax.plot(sx, sy2, 'k', linewidth=2)

        d = np.abs(np.nan_to_num(self.distances))
        colors = (d / d.max())
        area = 100.0 * np.power(1.0 + d, 2)

        sc = ax.scatter(self.sigma_re, self.sigma_im, c=colors, s=area, cmap='winter', alpha=0.75)

        annot = ax.annotate("", xy=(0, 0), xytext=(20, 20),
                            textcoords="offset points",
                            bbox=dict(boxstyle="round", fc="w"),
                            arrowprops=dict(arrowstyle="->"))
        annot.set_visible(False)

        ax.set_title('$\Sigma$ plot')
        ax.set_xlabel('$\sigma_{re}$')
        ax.set_ylabel('$\sigma_{im}$')

        def update_annotation(ind):
            pos = sc.get_offsets()[ind["ind"][0]]
            annot.xy = pos
            text = "{}".format("\n".join([names[n] for n in ind["ind"]]))
            annot.set_text(text)
            annot.get_bbox_patch().set_alpha(0.8)

        def hover(event):
            if event.inaxes == ax:
                cont, ind = sc.contains(event)
                if cont:
                    update_annotation(ind)
                    annot.set_visible(True)
                    fig.canvas.draw_idle()
                else:
                    if annot.get_visible():
                        annot.set_visible(False)
                        fig.canvas.draw_idle()

        fig.canvas.mpl_connect("motion_notify_event", hover)

    def mdl(self, result_type: ResultTypes, indices=None, names=None) -> "ResultsTable":
        """

        :param result_type:
        :param ax:
        :param indices:
        :param names:
        :return:
        """

        if indices is None and names is not None:
            indices = np.array(range(len(names)))

        if len(indices) > 0:
            labels = names[indices]

            if result_type == ResultTypes.SigmaDistances:
                y = np.abs(self.distances[indices])
                y_label = '(p.u.)'
                title = 'Sigma distances '

            elif result_type == ResultTypes.SigmaReal:
                y = self.sigma_re[indices]
                y_label = '(deg)'
                title = 'Real sigma '

            elif result_type == ResultTypes.SigmaImag:
                y = self.sigma_im[indices]
                y_label = '(p.u.)'
                title = 'Imaginary Sigma '

            elif result_type == ResultTypes.SigmaPlusDistances:
                y = np.c_[self.sigma_re[indices], self.sigma_im[indices], self.distances[indices]]
                y_label = '(p.u.)'
                title = 'Sigma and distances'

                mdl = ResultsTable(data=y, index=labels, columns=['σ real', 'σ imaginary', 'Distances'],
                                   title=title, ylabel=y_label, units=y_label)
                return mdl

            else:
                n = len(labels)
                y = np.zeros(n)
                y_label = ''
                title = ''

            # assemble model
            mdl = ResultsTable(data=y, index=labels, columns=[result_type.value[0]],
                               title=title, ylabel=y_label, units=y_label)
            return mdl

        else:
            return None


def multi_island_sigma(multi_circuit: MultiCircuit, options: PowerFlowOptions, logger=Logger()) -> "SigmaAnalysisResults":
    """
    Multiple islands power flow (this is the most generic power flow function)
    :param multi_circuit: MultiCircuit instance
    :param options: PowerFlowOptions instance
    :param logger: list of events to add to
    :return: PowerFlowResults instance
    """
    # print('PowerFlowDriver at ', self.grid.name)
    n = len(multi_circuit.buses)
    m = multi_circuit.get_branch_number()
    results = SigmaAnalysisResults(n)

    nc = compile_snapshot_circuit(circuit=multi_circuit,
                                  apply_temperature=options.apply_temperature_correction,
                                  branch_tolerance_mode=options.branch_impedance_tolerance_mode,
                                  opf_results=None)
    results.bus_names = nc.bus_data.bus_names

    calculation_inputs = nc.split_into_islands(ignore_single_node_islands=options.ignore_single_node_islands)

    if len(calculation_inputs) > 1:

        # simulate each island and merge the results
        for i, calculation_input in enumerate(calculation_inputs):

            if len(calculation_input.vd) > 0:
                # V, converged, norm_f, Scalc, iter_, elapsed, Sig_re, Sig_im
                U, X, Q, V, iter_ = helm_coefficients_josep(Ybus=calculation_input.Ybus,
                                                            Yseries=calculation_input.Yseries,
                                                            V0=calculation_input.Vbus,
                                                            S0=calculation_input.Sbus,
                                                            Ysh0=calculation_input.Yshunt,
                                                            pq=calculation_input.pq,
                                                            pv=calculation_input.pv,
                                                            sl=calculation_input.vd,
                                                            pqpv=calculation_input.pqpv,
                                                            tolerance=options.tolerance,
                                                            max_coeff=options.max_iter,
                                                            verbose=False,
                                                            logger=logger)

                # compute the sigma values
                n = calculation_input.nbus
                Sig_re = np.zeros(n, dtype=float)
                Sig_im = np.zeros(n, dtype=float)
                Sigma = sigma_function(U, X, iter_ - 1, calculation_input.Vbus[calculation_input.vd])
                Sig_re[calculation_input.pqpv] = np.real(Sigma)
                Sig_im[calculation_input.pqpv] = np.imag(Sigma)
                sigma_distances = sigma_distance(Sig_re, Sig_im)

                # store the results
                island_results = SigmaAnalysisResults(n=len(calculation_input.Vbus))
                island_results.lambda_value = 1.0
                island_results.Sbus = calculation_input.Sbus
                island_results.sigma_re = Sig_re
                island_results.sigma_im = Sig_im
                island_results.distances = sigma_distances

                bus_original_idx = calculation_input.original_bus_idx

                # merge the results from this island
                results.apply_from_island(island_results, bus_original_idx)

            else:
                logger.add_info('No slack nodes in the island', str(i))
    else:

        if len(calculation_inputs[0].vd) > 0:
            # only one island
            calculation_input = calculation_inputs[0]

            U, X, Q, V, iter_ = helm_coefficients_josep(Ybus=calculation_input.Ybus,
                                                        Yseries=calculation_input.Yseries,
                                                        V0=calculation_input.Vbus,
                                                        S0=calculation_input.Sbus,
                                                        Ysh0=calculation_input.Yshunt,
                                                        pq=calculation_input.pq,
                                                        pv=calculation_input.pv,
                                                        sl=calculation_input.vd,
                                                        pqpv=calculation_input.pqpv,
                                                        tolerance=options.tolerance,
                                                        max_coeff=options.max_iter,
                                                        verbose=False,
                                                        logger=logger)

            # compute the sigma values
            n = calculation_input.nbus
            Sig_re = np.zeros(n, dtype=float)
            Sig_im = np.zeros(n, dtype=float)
            Sigma = sigma_function(U, X, iter_ - 1, calculation_input.Vbus[calculation_input.vd])
            Sig_re[calculation_input.pqpv] = np.real(Sigma)
            Sig_im[calculation_input.pqpv] = np.imag(Sigma)
            sigma_distances = sigma_distance(Sig_re, Sig_im)

            # store the results
            island_results = SigmaAnalysisResults(n=len(calculation_input.Vbus))
            island_results.lambda_value = 1.0
            island_results.Sbus = calculation_input.Sbus
            island_results.sigma_re = Sig_re
            island_results.sigma_im = Sig_im
            island_results.distances = sigma_distances

            results.apply_from_island(island_results, calculation_input.original_bus_idx)
        else:
            logger.add_error('There are no slack nodes')

    return results


@nb.jit(cache=True)
def sigma_distance(sigma_real, sigma_imag):
    """
    Distance to the collapse in the sigma space

    The boundary curve is given by y = sqrt(1/4 + x)

    the distance is d = sqrt((x-a)^2 + (sqrt(1/4+ x) - b)^2)

    the derivative of this is d'=(2 (-a + x) + (-b + sqrt(1/4 + x))/sqrt(1/4 + x))/(2 sqrt((-a + x)^2 + (-b + sqrt(1/4 + x))^2))

    Making d'=0, and solving for x, we obtain:

    x1 = 1/12 (-64 a^3 + 48 a^2
               + 12 sqrt(3) sqrt(-64 a^3 b^2 + 48 a^2 b^2 - 12 a b^2 + 108 b^4 + b^2)
               - 12 a + 216 b^2 + 1)^(1/3) - (-256 a^2 + 128 a - 16)/
         (192 (-64 a^3 + 48 a^2
               + 12 sqrt(3) sqrt(-64 a^3 b^2 + 48 a^2 b^2 - 12 a b^2 + 108 b^4 + b^2)
               - 12 a + 216 b^2 + 1)^(1/3)) + 1/12 (8 a - 5)

    x2 = 1/12 (-64 a^3 + 48 a^2
               + 12 sqrt(3) sqrt(-64 a^3 b^2 + 48 a^2 b^2 - 12 a b^2 + 108 b^4 + b^2)
               - 12 a + 216 b^2 + 1)^(1/3) - (-256 a^2 + 128 a - 16) /
         (192 (-64 a^3 + 48 a^2
               + 12 sqrt(3) sqrt(-64 a^3 b^2 + 48 a^2 b^2 - 12 a b^2 + 108 b^4 + b^2)
               - 12 a + 216 b^2 + 1)^(1/3)) + 1/12 (8 a - 5)
    :param sigma_real: Sigma real array
    :param sigma_imag: Sigma imag array
    :return: distance of the sigma point to the curve sqrt(0.25 + x)
    """
    n = len(sigma_real)
    x1 = np.zeros(n)

    i = 0
    sq3 = np.sqrt(3)

    for a, b in zip(sigma_real, sigma_imag):

        t0 = -64 * a ** 3 * b ** 2 \
             + 48 * a ** 2 * b ** 2 \
             - 12 * a * b ** 2 \
             + 108 * b ** 4 + b ** 2

        if t0 > 0:

            t1 = (-64 * a**3
                  + 48 * a**2
                  + 12 * sq3 * np.sqrt(t0)
                  - 12 * a + 216 * b**2 + 1)**(1 / 3)

            # the value is within limits
            x1[i] = 1 / 12 * t1 - (-256 * a**2 + 128 * a - 16) / (192 * t1) + 1 / 12 * (8 * a - 5)
        else:
            t1 = (-64 * a ** 3
                  + 48 * a ** 2
                  + 12 * sq3 * np.sqrt(-t0)
                  - 12 * a + 216 * b ** 2 + 1) ** (1 / 3)

            # here I set the value negative to indicate that it is off-limits
            x1[i] = -(1 / 12 * t1 - (-256 * a**2 + 128 * a - 16) / (192 * t1) + 1 / 12 * (8 * a - 5))

        i += 1

    return x1


class SigmaAnalysisDriver(DriverTemplate):
    name = 'Sigma Analysis'
    tpe = SimulationTypes.SigmaAnalysis_run

    def __init__(self, grid: MultiCircuit, options: PowerFlowOptions):
        """
        PowerFlowDriver class constructor
        :param grid: MultiCircuit instance
        :param options: PowerFlowOptions instance
        """
        DriverTemplate.__init__(self, grid=grid)

        # Options to use
        self.options = options

        self.results = None

        self.logger = Logger()

        self.convergence_reports = list()

        self.__cancel__ = False

    @staticmethod
    def get_steps():
        """

        :return:
        """
        return list()

    def run(self):
        """
        Pack run_pf for the QThread
        :return:
        """
        self.results = multi_island_sigma(multi_circuit=self.grid,
                                          options=self.options,
                                          logger=self.logger)

    def cancel(self):
        self.__cancel__ = True




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import time
import scipy

from GridCal.Engine.Simulations.sparse_solve import get_sparse_type, get_linear_solver
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.ac_jacobian import AC_jacobian
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.common_functions import *
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults
from GridCal.Engine.basic_structures import ReactivePowerControlMode
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.discrete_controls import control_q_inside_method
from GridCal.Engine.basic_structures import Logger

linear_solver = get_linear_solver()
sparse = get_sparse_type()
scipy.ALLOW_THREADS = True
np.set_printoptions(precision=8, suppress=True, linewidth=320)


def levenberg_marquardt_pf(Ybus, S0, V0, I0, Y0, pv_, pq_, Qmin, Qmax, tol, max_it=50,
                           control_q=ReactivePowerControlMode.NoControl,
                           verbose=False, logger: Logger = None) -> NumericPowerFlowResults:
    """
    Solves the power flow problem by the Levenberg-Marquardt power flow algorithm.
    It is usually better than Newton-Raphson, but it takes an order of magnitude more time to converge.
    @Author: Santiago Peñate Vera
    :param Ybus: Admittance matrix
    :param S0: Array of nodal power injections (ZIP)
    :param V0: Array of nodal voltages (initial solution)
    :param I0: Array of nodal current injections (ZIP)
    :param Y0: Array of nodal admittance injections (ZIP)
    :param pv_: Array with the indices of the PV buses
    :param pq_: Array with the indices of the PQ buses
    :param Qmin: array of lower reactive power limits per bus
    :param Qmax: array of upper reactive power limits per bus
    :param tol: Tolerance
    :param max_it: Maximum number of iterations
    :param control_q: Type of reactive power control
    :param verbose: Display console information
    :param logger: Logger instance
    :return: NumericPowerFlowResults instance
    """
    start = time.time()

    # initialize
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pq = pq_.copy()
    pv = pv_.copy()
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)
    npvpq = npq + npv

    if npvpq > 0:
        normF = 100000
        update_jacobian = True
        converged = False
        iter_ = 0
        nu = 2.0
        lbmda = 0
        f_prev = 1e9  # very large number
        nn = 2 * npq + npv
        Idn = sp.diags(np.ones(nn))  # csc_matrix identity
        H: sp.csc_matrix = sp.csc_matrix((0, 0))
        H1: sp.csc_matrix = sp.csc_matrix((0, 0))
        H2: sp.csc_matrix = sp.csc_matrix((0, 0))
        Scalc = S0  # is updated later

        if verbose > 1:
            logger.add_debug('LM previous values ' + '-' * 200)
            logger.add_debug('Y (real):\n', Ybus.real.toarray())
            logger.add_debug('Y (imag):\n', Ybus.imag.toarray())
            logger.add_debug('pvpq:\n', pvpq)
            logger.add_debug('pq:\n', pq)
            logger.add_debug('Vm:\n', Vm)
            logger.add_debug('Va:\n', Va)

        while not converged and iter_ < max_it:

            if verbose:
                logger.add_debug('LM Iteration {0}'.format(iter_) + '-' * 200)

            # evaluate Jacobian
            if update_jacobian:
                H = AC_jacobian(Ybus, V, pvpq, pq, npv, npq)

                # system matrix
                # H1 = H^t
                H1 = H.transpose()  # .tocsr()

                # H2 = H1·H
                H2 = H1.dot(H)

            # evaluate the solution error F(x0)
            Sbus = compute_zip_power(S0, I0, Y0, Vm)
            Scalc = compute_power(Ybus, V)
            dz = compute_fx(Scalc, Sbus, pvpq, pq)

            # set first value of lmbda
            if iter_ == 0:
                lbmda = 1e-3 * H2.diagonal().max()

            # compute system matrix A = H^T·H - lambda·I
            A = H2 + lbmda * Idn

            # right-hand side
            # H^t·dz
            rhs = H1.dot(dz)

            # Solve the increment
            dx = linear_solver(A, rhs)

            # objective function to minimize
            f = 0.5 * dz.dot(dz)

            # decision function
            val = dx.dot(lbmda * dx + rhs)
            if val > 0.0:
                rho = (f_prev - f) / (0.5 * val)
            else:
                rho = -1.0

            if verbose > 1:
                logger.add_debug('rho:', rho)

            # lambda update
            if rho >= 0:
                update_jacobian = True
                lbmda *= max([1.0 / 3.0, 1 - (2 * rho - 1) ** 3])
                nu = 2.0

                # reassign the solution vector
                dVa[pvpq] = dx[:npvpq]
                dVm[pq] = dx[npvpq:]

                # update Vm and Va again in case we wrapped around with a negative Vm
                Vm -= dVm
                Va -= dVa
                V = polar_to_rect(Vm, Va)

                if verbose > 1:
                    logger.add_debug('J:\n', H.toarray())
                    logger.add_debug('f:\n', rhs)
                    logger.add_debug('Vm:\n', Vm)
                    logger.add_debug('Va:\n', Va)

            else:
                update_jacobian = False
                lbmda *= nu
                nu *= 2.0

            # check convergence
            Sbus = compute_zip_power(S0, I0, Y0, Vm)
            Scalc = compute_power(Ybus, V)
            e = compute_fx(Scalc, Sbus, pvpq, pq)
            normF = compute_fx_error(e)

            if verbose:
                logger.add_debug('error', normF)

            # review reactive power limits
            # it is only worth checking Q limits with a low error
            # since with higher errors, the Q values may be far from realistic
            # finally, the Q control only makes sense if there are pv nodes
            if control_q != ReactivePowerControlMode.NoControl and normF < 1e-2 and npv > 0:

                # check and adjust the reactive power
                # this function passes pv buses to pq when the limits are violated,
                # but not pq to pv because that is unstable
                n_changes, Scalc, Sbus, pv, pq, pvpq, messages = control_q_inside_method(Scalc, Sbus, pv, pq,
                                                                                         pvpq, Qmin, Qmax)

                if n_changes > 0:
                    # adjust internal variables to the new pq|pv values
                    npv = len(pv)
                    npq = len(pq)
                    npvpq = npv + npq
                    pvpq_lookup = np.zeros(Ybus.shape[0], dtype=int)
                    pvpq_lookup[pvpq] = np.arange(npvpq)

                    nn = 2 * npq + npv
                    ii = np.linspace(0, nn - 1, nn)
                    Idn = sparse((np.ones(nn), (ii, ii)), shape=(nn, nn))  # csc_matrix identity

                    # recompute the error based on the new Sbus
                    e = compute_fx(Scalc, Sbus, pvpq, pq)
                    normF = compute_fx_error(e)

                    if verbose > 0:
                        for sense, idx, var in messages:
                            msg = "Bus " + str(idx) + " changed to PQ, limited to " + str(var * 100) + " MVAr"
                            logger.add_debug(msg)

            converged = normF < tol
            f_prev = f

            # update iteration counter
            iter_ += 1
    else:
        normF = 0
        converged = True
        Scalc = compute_zip_power(S0, I0, Y0, Vm)  # compute the ZIP power injection
        iter_ = 0

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, converged, normF, Scalc, None, None, None, None, None, None, iter_, elapsed)




# -*- coding: utf-8 -*-

# Resource object code
#
# Created: lun. ene. 6 14:16:48 2020
#      by: The Resource Compiler for PySide2 (Qt v5.13.0)
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore

qt_resource_data = b"\
\x00\x00\x07\x02\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<svg\x0a   xm\
lns:dc=\x22http://p\
url.org/dc/eleme\
nts/1.1/\x22\x0a   xml\
ns:cc=\x22http://cr\
eativecommons.or\
g/ns#\x22\x0a   xmlns:\
rdf=\x22http://www.\
w3.org/1999/02/2\
2-rdf-syntax-ns#\
\x22\x0a   xmlns:svg=\x22\
http://www.w3.or\
g/2000/svg\x22\x0a   x\
mlns=\x22http://www\
.w3.org/2000/svg\
\x22\x0a   xmlns:sodip\
odi=\x22http://sodi\
podi.sourceforge\
.net/DTD/sodipod\
i-0.dtd\x22\x0a   xmln\
s:inkscape=\x22http\
://www.inkscape.\
org/namespaces/i\
nkscape\x22\x0a   widt\
h=\x2248\x22\x0a   height\
=\x2248\x22\x0a   viewBox\
=\x220 0 48 48\x22\x0a   \
version=\x221.1\x22\x0a  \
 id=\x22svg4\x22\x0a   so\
dipodi:docname=\x22\
accept.svg\x22\x0a   i\
nkscape:version=\
\x220.92.3 (2405546\
, 2018-03-11)\x22>\x0a\
  <metadata\x0a    \
 id=\x22metadata10\x22\
>\x0a    <rdf:RDF>\x0a\
      <cc:Work\x0a \
        rdf:abou\
t=\x22\x22>\x0a        <d\
c:format>image/s\
vg+xml</dc:forma\
t>\x0a        <dc:t\
ype\x0a           r\
df:resource=\x22htt\
p://purl.org/dc/\
dcmitype/StillIm\
age\x22 />\x0a      </\
cc:Work>\x0a    </r\
df:RDF>\x0a  </meta\
data>\x0a  <defs\x0a  \
   id=\x22defs8\x22 />\
\x0a  <sodipodi:nam\
edview\x0a     page\
color=\x22#ffffff\x22\x0a\
     bordercolor\
=\x22#666666\x22\x0a     \
borderopacity=\x221\
\x22\x0a     objecttol\
erance=\x2210\x22\x0a    \
 gridtolerance=\x22\
10\x22\x0a     guideto\
lerance=\x2210\x22\x0a   \
  inkscape:pageo\
pacity=\x220\x22\x0a     \
inkscape:pagesha\
dow=\x222\x22\x0a     ink\
scape:window-wid\
th=\x221853\x22\x0a     i\
nkscape:window-h\
eight=\x221025\x22\x0a   \
  id=\x22namedview6\
\x22\x0a     showgrid=\
\x22false\x22\x0a     ink\
scape:zoom=\x222.45\
83333\x22\x0a     inks\
cape:cx=\x22-184.46\
43\x22\x0a     inkscap\
e:cy=\x2239.302561\x22\
\x0a     inkscape:w\
indow-x=\x2267\x22\x0a   \
  inkscape:windo\
w-y=\x2227\x22\x0a     in\
kscape:window-ma\
ximized=\x221\x22\x0a    \
 inkscape:curren\
t-layer=\x22svg4\x22 /\
>\x0a  <path\x0a     s\
tyle=\x22fill:#00d4\
aa;fill-rule:eve\
nodd;stroke:#00d\
4aa;stroke-width\
:1.15240359px;st\
roke-linecap:but\
t;stroke-linejoi\
n:miter;stroke-o\
pacity:1\x22\x0a     d\
=\x22M 4.1877492,25\
.155303 15.67271\
9,39.687308 45.9\
08665,13.904718 \
41.455309,10.857\
686 17.079041,29\
.139883 8.875491\
5,23.748979 Z\x22\x0a \
    id=\x22path815\x22\
\x0a     inkscape:c\
onnector-curvatu\
re=\x220\x22 />\x0a</svg>\
\x0a\
\x00\x00\x0b\xd0\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:sodipodi=\x22\
http://sodipodi.\
sourceforge.net/\
DTD/sodipodi-0.d\
td\x22\x0a   xmlns:ink\
scape=\x22http://ww\
w.inkscape.org/n\
amespaces/inksca\
pe\x22\x0a   version=\x22\
1.1\x22\x0a   id=\x22svg2\
\x22\x0a   width=\x22192\x22\
\x0a   height=\x22192\x22\
\x0a   viewBox=\x220 0\
 192 192\x22\x0a   sod\
ipodi:docname=\x22d\
elete2.svg\x22\x0a   i\
nkscape:version=\
\x220.92.3 (2405546\
, 2018-03-11)\x22>\x0a\
  <metadata\x0a    \
 id=\x22metadata8\x22>\
\x0a    <rdf:RDF>\x0a \
     <cc:Work\x0a  \
       rdf:about\
=\x22\x22>\x0a        <dc\
:format>image/sv\
g+xml</dc:format\
>\x0a        <dc:ty\
pe\x0a           rd\
f:resource=\x22http\
://purl.org/dc/d\
cmitype/StillIma\
ge\x22 />\x0a        <\
dc:title />\x0a    \
  </cc:Work>\x0a   \
 </rdf:RDF>\x0a  </\
metadata>\x0a  <def\
s\x0a     id=\x22defs6\
\x22 />\x0a  <sodipodi\
:namedview\x0a     \
pagecolor=\x22#ffff\
ff\x22\x0a     borderc\
olor=\x22#666666\x22\x0a \
    borderopacit\
y=\x221\x22\x0a     objec\
ttolerance=\x2210\x22\x0a\
     gridtoleran\
ce=\x2210\x22\x0a     gui\
detolerance=\x2210\x22\
\x0a     inkscape:p\
ageopacity=\x220\x22\x0a \
    inkscape:pag\
eshadow=\x222\x22\x0a    \
 inkscape:window\
-width=\x221853\x22\x0a  \
   inkscape:wind\
ow-height=\x221025\x22\
\x0a     id=\x22namedv\
iew4\x22\x0a     showg\
rid=\x22false\x22\x0a    \
 inkscape:zoom=\x22\
1.7383042\x22\x0a     \
inkscape:cx=\x22-10\
5.16921\x22\x0a     in\
kscape:cy=\x226.575\
1461\x22\x0a     inksc\
ape:window-x=\x2267\
\x22\x0a     inkscape:\
window-y=\x2227\x22\x0a  \
   inkscape:wind\
ow-maximized=\x221\x22\
\x0a     inkscape:c\
urrent-layer=\x22g8\
31\x22 />\x0a  <g\x0a    \
 id=\x22g831\x22\x0a     \
style=\x22stroke:#3\
7c8ab;stroke-wid\
th:21;stroke-mit\
erlimit:4;stroke\
-dasharray:none\x22\
>\x0a    <path\x0a    \
   inkscape:conn\
ector-curvature=\
\x220\x22\x0a       id=\x22p\
ath812\x22\x0a       d\
=\x22M 58.686516,34\
.154113 133.6998\
2,108.38751\x22\x0a   \
    style=\x22fill:\
none;fill-rule:e\
venodd;stroke:#f\
f8080;stroke-wid\
th:15.57077312;s\
troke-linecap:ro\
und;stroke-linej\
oin:miter;stroke\
-miterlimit:4;st\
roke-dasharray:n\
one;stroke-opaci\
ty:1\x22\x0a       sod\
ipodi:nodetypes=\
\x22cc\x22 />\x0a    <pat\
h\x0a       inkscap\
e:connector-curv\
ature=\x220\x22\x0a      \
 id=\x22path812-3\x22\x0a\
       d=\x22M 132.\
74009,34.474021 \
59.646243,108.06\
76\x22\x0a       style\
=\x22fill:none;fill\
-rule:evenodd;st\
roke:#ff8080;str\
oke-width:15.570\
77312;stroke-lin\
ecap:round;strok\
e-linejoin:miter\
;stroke-miterlim\
it:4;stroke-dash\
array:none;strok\
e-opacity:1\x22\x0a   \
    sodipodi:nod\
etypes=\x22cc\x22 />\x0a \
   <path\x0a       \
inkscape:connect\
or-curvature=\x220\x22\
\x0a       id=\x22path\
812-7\x22\x0a       d=\
\x22M 48.209069,89.\
302145 123.22237\
,163.53554\x22\x0a    \
   style=\x22fill:n\
one;fill-rule:ev\
enodd;stroke:#ff\
8080;stroke-widt\
h:15.57077312;st\
roke-linecap:rou\
nd;stroke-linejo\
in:miter;stroke-\
miterlimit:4;str\
oke-dasharray:no\
ne;stroke-opacit\
y:1\x22\x0a       sodi\
podi:nodetypes=\x22\
cc\x22 />\x0a    <path\
\x0a       inkscape\
:connector-curva\
ture=\x220\x22\x0a       \
id=\x22path812-3-5\x22\
\x0a       d=\x22M 122\
.26264,89.622055\
 49.168796,163.2\
1563\x22\x0a       sty\
le=\x22fill:none;fi\
ll-rule:evenodd;\
stroke:#ff8080;s\
troke-width:15.5\
7077312;stroke-l\
inecap:round;str\
oke-linejoin:mit\
er;stroke-miterl\
imit:4;stroke-da\
sharray:none;str\
oke-opacity:1\x22\x0a \
      sodipodi:n\
odetypes=\x22cc\x22 />\
\x0a  </g>\x0a</svg>\x0a\
\x00\x00\x0ai\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<svg\x0a   xm\
lns:dc=\x22http://p\
url.org/dc/eleme\
nts/1.1/\x22\x0a   xml\
ns:cc=\x22http://cr\
eativecommons.or\
g/ns#\x22\x0a   xmlns:\
rdf=\x22http://www.\
w3.org/1999/02/2\
2-rdf-syntax-ns#\
\x22\x0a   xmlns:svg=\x22\
http://www.w3.or\
g/2000/svg\x22\x0a   x\
mlns=\x22http://www\
.w3.org/2000/svg\
\x22\x0a   xmlns:sodip\
odi=\x22http://sodi\
podi.sourceforge\
.net/DTD/sodipod\
i-0.dtd\x22\x0a   xmln\
s:inkscape=\x22http\
://www.inkscape.\
org/namespaces/i\
nkscape\x22\x0a   widt\
h=\x2248\x22\x0a   height\
=\x2248\x22\x0a   viewBox\
=\x220 0 48 48\x22\x0a   \
version=\x221.1\x22\x0a  \
 id=\x22svg4\x22\x0a   so\
dipodi:docname=\x22\
accept2.svg\x22\x0a   \
inkscape:version\
=\x220.92.3 (240554\
6, 2018-03-11)\x22>\
\x0a  <metadata\x0a   \
  id=\x22metadata10\
\x22>\x0a    <rdf:RDF>\
\x0a      <cc:Work\x0a\
         rdf:abo\
ut=\x22\x22>\x0a        <\
dc:format>image/\
svg+xml</dc:form\
at>\x0a        <dc:\
type\x0a           \
rdf:resource=\x22ht\
tp://purl.org/dc\
/dcmitype/StillI\
mage\x22 />\x0a      <\
/cc:Work>\x0a    </\
rdf:RDF>\x0a  </met\
adata>\x0a  <defs\x0a \
    id=\x22defs8\x22 /\
>\x0a  <sodipodi:na\
medview\x0a     pag\
ecolor=\x22#ffffff\x22\
\x0a     bordercolo\
r=\x22#666666\x22\x0a    \
 borderopacity=\x22\
1\x22\x0a     objectto\
lerance=\x2210\x22\x0a   \
  gridtolerance=\
\x2210\x22\x0a     guidet\
olerance=\x2210\x22\x0a  \
   inkscape:page\
opacity=\x220\x22\x0a    \
 inkscape:pagesh\
adow=\x222\x22\x0a     in\
kscape:window-wi\
dth=\x221853\x22\x0a     \
inkscape:window-\
height=\x221025\x22\x0a  \
   id=\x22namedview\
6\x22\x0a     showgrid\
=\x22false\x22\x0a     in\
kscape:zoom=\x226.3\
333333\x22\x0a     ink\
scape:cx=\x2226.339\
228\x22\x0a     inksca\
pe:cy=\x2250.577833\
\x22\x0a     inkscape:\
window-x=\x2267\x22\x0a  \
   inkscape:wind\
ow-y=\x2227\x22\x0a     i\
nkscape:window-m\
aximized=\x221\x22\x0a   \
  inkscape:curre\
nt-layer=\x22svg4\x22 \
/>\x0a  <path\x0a     \
style=\x22fill:#00d\
4aa;fill-rule:ev\
enodd;stroke:#00\
d4aa;stroke-widt\
h:0.73550624px;s\
troke-linecap:bu\
tt;stroke-linejo\
in:miter;stroke-\
opacity:1\x22\x0a     \
d=\x22M 2.2550022,3\
5.606185 9.58513\
14,44.881044 28.\
882821,28.425649\
 26.040526,26.48\
0922 10.482698,3\
8.14929 5.246891\
2,34.708618 Z\x22\x0a \
    id=\x22path815\x22\
\x0a     inkscape:c\
onnector-curvatu\
re=\x220\x22 />\x0a  <g\x0a \
    style=\x22strok\
e:#37c8ab;stroke\
-width:21;stroke\
-miterlimit:4;st\
roke-dasharray:n\
one\x22\x0a     id=\x22g8\
58\x22\x0a     transfo\
rm=\x22matrix(0.225\
97714,0,0,0.2259\
7714,-8.857672,-\
3.3950478)\x22>\x0a   \
 <path\x0a       so\
dipodi:nodetypes\
=\x22cc\x22\x0a       sty\
le=\x22fill:none;fi\
ll-rule:evenodd;\
stroke:#ff8080;s\
troke-width:15.5\
7077312;stroke-l\
inecap:round;str\
oke-linejoin:mit\
er;stroke-miterl\
imit:4;stroke-da\
sharray:none;str\
oke-opacity:1\x22\x0a \
      d=\x22M 58.68\
6516,34.154113 1\
33.69982,108.387\
51\x22\x0a       id=\x22p\
ath812\x22\x0a       i\
nkscape:connecto\
r-curvature=\x220\x22 \
/>\x0a    <path\x0a   \
    sodipodi:nod\
etypes=\x22cc\x22\x0a    \
   style=\x22fill:n\
one;fill-rule:ev\
enodd;stroke:#ff\
8080;stroke-widt\
h:15.57077312;st\
roke-linecap:rou\
nd;stroke-linejo\
in:miter;stroke-\
miterlimit:4;str\
oke-dasharray:no\
ne;stroke-opacit\
y:1\x22\x0a       d=\x22M\
 132.74009,34.47\
4021 59.646243,1\
08.0676\x22\x0a       \
id=\x22path812-3\x22\x0a \
      inkscape:c\
onnector-curvatu\
re=\x220\x22 />\x0a  </g>\
\x0a</svg>\x0a\
\x00\x00\x09\x10\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:sodipodi=\x22\
http://sodipodi.\
sourceforge.net/\
DTD/sodipodi-0.d\
td\x22\x0a   xmlns:ink\
scape=\x22http://ww\
w.inkscape.org/n\
amespaces/inksca\
pe\x22\x0a   version=\x22\
1.1\x22\x0a   id=\x22svg2\
\x22\x0a   width=\x22192\x22\
\x0a   height=\x22192\x22\
\x0a   viewBox=\x220 0\
 192 192\x22\x0a   sod\
ipodi:docname=\x22d\
elete3.svg\x22\x0a   i\
nkscape:version=\
\x220.92.4 (unknown\
)\x22>\x0a  <metadata\x0a\
     id=\x22metadat\
a8\x22>\x0a    <rdf:RD\
F>\x0a      <cc:Wor\
k\x0a         rdf:a\
bout=\x22\x22>\x0a       \
 <dc:format>imag\
e/svg+xml</dc:fo\
rmat>\x0a        <d\
c:type\x0a         \
  rdf:resource=\x22\
http://purl.org/\
dc/dcmitype/Stil\
lImage\x22 />\x0a     \
   <dc:title />\x0a\
      </cc:Work>\
\x0a    </rdf:RDF>\x0a\
  </metadata>\x0a  \
<defs\x0a     id=\x22d\
efs6\x22 />\x0a  <sodi\
podi:namedview\x0a \
    pagecolor=\x22#\
ffffff\x22\x0a     bor\
dercolor=\x22#66666\
6\x22\x0a     borderop\
acity=\x221\x22\x0a     o\
bjecttolerance=\x22\
10\x22\x0a     gridtol\
erance=\x2210\x22\x0a    \
 guidetolerance=\
\x2210\x22\x0a     inksca\
pe:pageopacity=\x22\
0\x22\x0a     inkscape\
:pageshadow=\x222\x22\x0a\
     inkscape:wi\
ndow-width=\x221863\
\x22\x0a     inkscape:\
window-height=\x221\
025\x22\x0a     id=\x22na\
medview4\x22\x0a     s\
howgrid=\x22false\x22\x0a\
     inkscape:zo\
om=\x221.7383042\x22\x0a \
    inkscape:cx=\
\x2228.86947\x22\x0a     \
inkscape:cy=\x226.5\
751461\x22\x0a     ink\
scape:window-x=\x22\
57\x22\x0a     inkscap\
e:window-y=\x2227\x22\x0a\
     inkscape:wi\
ndow-maximized=\x22\
1\x22\x0a     inkscape\
:current-layer=\x22\
g831\x22 />\x0a  <g\x0a  \
   id=\x22g831\x22\x0a   \
  style=\x22stroke:\
#37c8ab;stroke-w\
idth:21;stroke-m\
iterlimit:4;stro\
ke-dasharray:non\
e\x22>\x0a    <path\x0a  \
     inkscape:co\
nnector-curvatur\
e=\x220\x22\x0a       id=\
\x22path812\x22\x0a      \
 d=\x22M 59.837063,\
61.191958 134.85\
037,135.42536\x22\x0a \
      style=\x22fil\
l:none;fill-rule\
:evenodd;stroke:\
#ff8080;stroke-w\
idth:15.57077312\
;stroke-linecap:\
round;stroke-lin\
ejoin:miter;stro\
ke-miterlimit:4;\
stroke-dasharray\
:none;stroke-opa\
city:1\x22\x0a       s\
odipodi:nodetype\
s=\x22cc\x22 />\x0a    <p\
ath\x0a       inksc\
ape:connector-cu\
rvature=\x220\x22\x0a    \
   id=\x22path812-3\
\x22\x0a       d=\x22M 13\
3.89064,61.51186\
6 60.79679,135.1\
0545\x22\x0a       sty\
le=\x22fill:none;fi\
ll-rule:evenodd;\
stroke:#ff8080;s\
troke-width:15.5\
7077312;stroke-l\
inecap:round;str\
oke-linejoin:mit\
er;stroke-miterl\
imit:4;stroke-da\
sharray:none;str\
oke-opacity:1\x22\x0a \
      sodipodi:n\
odetypes=\x22cc\x22 />\
\x0a  </g>\x0a</svg>\x0a\
"

qt_resource_name = b"\
\x00\x05\
\x00O\xa6S\
\x00I\
\x00c\x00o\x00n\x00s\
\x00\x05\
\x00o\xa6S\
\x00i\
\x00c\x00o\x00n\x00s\
\x00\x0a\
\x0c{\xa9\xe7\
\x00a\
\x00c\x00c\x00e\x00p\x00t\x00.\x00s\x00v\x00g\
\x00\x0b\
\x0aP\xdfG\
\x00d\
\x00e\x00l\x00e\x00t\x00e\x002\x00.\x00s\x00v\x00g\
\x00\x0b\
\x07\xbaiG\
\x00a\
\x00c\x00c\x00e\x00p\x00t\x002\x00.\x00s\x00v\x00g\
\x00\x0a\
\x0c\xad\x02\x87\
\x00d\
\x00e\x00l\x00e\x00t\x00e\x00.\x00s\x00v\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x10\x00\x02\x00\x00\x00\x04\x00\x00\x00\x03\
\x00\x00\x00V\x00\x00\x00\x00\x00\x01\x00\x00\x12\xda\
\x00\x00\x00:\x00\x00\x00\x00\x00\x01\x00\x00\x07\x06\
\x00\x00\x00 \x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x00r\x00\x00\x00\x00\x00\x01\x00\x00\x1dG\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x01, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x01, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import pandas as pd
import sqlite3
import numpy as np

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.IO.excel_interface import check_names
from GridCal.Engine.IO.generic_io_functions import parse_config_df


def save_data_frames_to_sqlite(dfs, file_path, text_func=None, progress_func=None):
    """
    Save the circuit information in excel format
    :param dfs: list of DataFrames
    :param file_path: path to the excel file
    :return: logger with information
    """
    logger = Logger()

    conn = sqlite3.connect(file_path)

    n = len(dfs.keys())

    for i, key in enumerate(dfs.keys()):

        if progress_func is not None:
            progress_func((i + 1) / n * 100)

        if text_func is not None:
            text_func('Saving ' + key)

        dfs[key].to_sql(key, conn, if_exists='replace', index=False)

    return logger


def open_data_frames_from_sqlite(file_path, text_func=None, progress_func=None):

    # make connection
    conn = sqlite3.connect(file_path)

    dfs = dict()

    # get the table names
    tables = conn.execute("SELECT name FROM sqlite_master WHERE type='table';")

    names = [t[0] for t in tables]

    check_names(names)
    n = len(names)
    for i, key in enumerate(names):

        if progress_func is not None:
            progress_func((i + 1) / n * 100)

        if text_func is not None:
            text_func('select * from ' + key)

        dfs[key] = pd.read_sql('select * from ' + key, conn)

    # parse the configuration
    dfs = parse_config_df(dfs['config'], dfs)

    return dfs


if __name__ == '__main__':
    import time
    from GridCal.Engine.IO.file_handler import *
    from GridCal.Engine.IO.pack_unpack import create_data_frames, data_frames_to_circuit

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39.gridcal'

    a = time.time()
    circuit_ = FileOpen(fname).open()
    print('native based open:', time.time() - a)

    print('Saving .sqlite ...')
    dfs = dfs = create_data_frames(circuit=circuit_)
    save_data_frames_to_sqlite(dfs, file_path=circuit_.name + '.sqlite')

    a = time.time()
    data = open_data_frames_from_sqlite(circuit_.name + '.sqlite')
    circuit2 = data_frames_to_circuit(data)
    print('sql based open:', time.time() - a)
